# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import abstractmethod
import json
import os
from typing import Any, Iterator, List, Tuple, Optional, Dict, Union, Container
import pyproj
from shapely import Point
import xarray as xr
import xvec

from xcube.core.normalize import normalize_dataset
from xcube.core.store import GEO_DATA_FRAME_TYPE
from xcube.core.store import DATASET_TYPE
from xcube.core.store import DataDescriptor
from xcube.core.store import DataOpener
from xcube.core.store import DataStore
from xcube.core.store import DataStoreError
from xcube.core.store import DataType
from xcube.core.store import DatasetDescriptor
from xcube.core.store import GeoDataFrameDescriptor
from xcube.core.store import VariableDescriptor
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonDateSchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonNumberSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

from .cciodp import CciOdp
from .chunkstore import CciChunkStore
from .constants import CCI_ODD_URL
from .constants import DATAFRAME_OPENER_ID
from .constants import DATASET_OPENER_ID
from .constants import DEFAULT_NUM_RETRIES
from .constants import DEFAULT_RETRY_BACKOFF_BASE
from .constants import DEFAULT_RETRY_BACKOFF_MAX
from .constants import OPENSEARCH_CEDA_URL
from .constants import VECTORDATACUBE_OPENER_ID
from .dataframeaccess import DataFrameAccessor
from .normalize import normalize_coord_names
from .normalize import normalize_dims_description
from .normalize import normalize_variable_dims_description
from .normalize import normalize_var_infos
from .vdcaccess import VectorDataCubeDescriptor
from .vdcaccess import VECTOR_DATA_CUBE_TYPE

_RELEVANT_METADATA_ATTRIBUTES = \
    ['ecv', 'institute', 'processing_level', 'product_string',
     'product_version', 'data_type', 'abstract', 'title', 'licences',
     'publication_date', 'catalog_url', 'sensor_id', 'platform_id',
     'cci_project', 'description', 'project', 'references', 'source',
     'history', 'comment', 'uuid']
_INSTITUTES = ['Alfred-Wegener-Institut Helmholtz-Zentrum für '
               'Polar- und Meeresforschung', 'Plymouth Marine Laboratory',
               'ENVironmental Earth Observation IT GmbH',
               'multi-institution', 'DTU Space', 'Freie Universitaet Berlin',
               'Vienna University of Technology', 'Deutscher Wetterdienst',
               'Netherlands Institute for Space Research',
               'Technische Universität Dresden',
               'Institute of Environmental Physics',
               'Rutherford Appleton Laboratory',
               'Universite Catholique de Louvain', 'University of Alcala',
               'University of Leicester', 'Norwegian Meteorological Institute',
               'University of Bremen', 'Belgian Institute for Space Aeronomy',
               'Deutsches Zentrum fuer Luft- und Raumfahrt',
               'Royal Netherlands Meteorological Institute',
               'The Geological Survey of Denmark and Greenland']


def get_temporal_resolution_from_id(data_id: str) -> Optional[str]:
    data_time_res = data_id.split('.')[2]
    time_res_items = dict(D=['days', 'day'],
                          M=['months', 'mon', 'climatology'],
                          Y=['yrs', 'yr', 'year'])
    for time_res_pandas_id, time_res_ids_list in time_res_items.items():
        for i, time_res_id in enumerate(time_res_ids_list):
            if time_res_id in data_time_res:
                if i == 0:
                    return f'{data_time_res.split("-")[0]}{time_res_pandas_id}'
                return f'1{time_res_pandas_id}'


class CciOdpDataOpener(DataOpener):

    # noinspection PyShadowingBuiltins
    def __init__(self,
                 cci_cdc: CciOdp,
                 id: str,
                 data_type: DataType,
                 normalize_data: bool = True):
        self._cci_cdc = cci_cdc
        self._id = id
        self._data_type = data_type
        self._normalize_data = normalize_data

    @property
    def dataset_names(self) -> List[str]:
        return self._cci_cdc.dataset_names

    @classmethod
    @abstractmethod
    def get_data_types(cls) -> Tuple[DataType, ...]:
        pass

    def has_data(self, data_id: str) -> bool:
        return data_id in self.dataset_names

    def _describe_data(self, data_ids: List[str]) -> List[DatasetDescriptor]:
        ds_metadata_list = self._cci_cdc.get_datasets_metadata(data_ids)
        data_descriptors = []
        for i, ds_metadata in enumerate(ds_metadata_list):
            data_descriptors.append(self._get_data_descriptor_from_metadata(
                data_ids[i], ds_metadata)
            )
        return data_descriptors

    def describe_data(self, data_id: str) -> DataDescriptor:
        self._assert_valid_data_id(data_id)
        try:
            ds_metadata = self._cci_cdc.get_dataset_metadata(data_id)
            return self._get_data_descriptor_from_metadata(data_id, ds_metadata)
        except ValueError:
            raise DataStoreError(
                f'Cannot describe metadata. '
                f'"{data_id}" does not seem to be a valid identifier.'
            )

    # noinspection PyArgumentList
    @abstractmethod
    def _get_data_descriptor_from_metadata(self,
                                           data_id: str,
                                           metadata: dict) -> DatasetDescriptor:
        pass

    def _get_variable_descriptors(self,
                                  var_names: List[str],
                                  var_infos: dict,
                                  time_dim_name: str,
                                  normalize_dims: bool = True) \
            -> Dict[str, VariableDescriptor]:
        var_descriptors = {}
        for var_name in var_names:
            if var_name in var_infos:
                var_info = var_infos[var_name]
                var_dtype = var_info['data_type']
                var_dims = self._normalize_var_dims(
                    var_info['dimensions'], time_dim_name) \
                    if normalize_dims else var_info['dimensions']
                var_descriptors[var_name] = VariableDescriptor(
                    var_name, var_dtype, var_dims, attrs=var_info
                )
            else:
                var_descriptors[var_name] = VariableDescriptor(
                    var_name, '', ''
                )
        return var_descriptors

    @staticmethod
    def _remove_irrelevant_metadata_attributes(attrs: dict):
        to_remove_list = []
        for attribute in attrs:
            if attribute not in _RELEVANT_METADATA_ATTRIBUTES:
                to_remove_list.append(attribute)
        for to_remove in to_remove_list:
            attrs.pop(to_remove)

    def search_data(self, **search_params) -> Iterator[DatasetDescriptor]:
        search_result = self._cci_cdc.search(**search_params)
        data_descriptors = self._describe_data(search_result)
        return iter(data_descriptors)

    def get_open_data_params_schema(
            self, data_id: str = None
    ) -> JsonObjectSchema:
        if data_id is None:
            return self._get_open_data_params_schema()
        self._assert_valid_data_id(data_id)
        dsd = self.describe_data(data_id)
        return self._get_open_data_params_schema(dsd)

    @abstractmethod
    def _get_open_data_params_schema(
            self,
            dsd: DatasetDescriptor = None
    ) -> JsonObjectSchema:
        pass

    @abstractmethod
    def open_data(self, data_id: str, **open_params) -> Any:
        pass

    def _assert_valid_data_id(self, data_id: str):
        if data_id not in self.dataset_names:
            raise DataStoreError(f'Cannot describe metadata of '
                                 f'data resource "{data_id}", '
                                 f'as it cannot be accessed by '
                                 f'data accessor "{self._id}".')

    def _normalize_dataset(self, ds: xr.Dataset) -> xr.Dataset:
        if self._normalize_data:
            return normalize_dataset(ds)
        return ds

    def _normalize_dims(self, dims: dict) -> dict:
        if self._normalize_data:
            return normalize_dims_description(dims)
        return dims.copy()

    def _normalize_var_dims(self, var_dims: List[str], time_dim_name: str) \
            -> Optional[List[str]]:
        if self._normalize_data:
            return normalize_variable_dims_description(var_dims)
        new_var_dims = var_dims.copy()
        if time_dim_name not in new_var_dims and len(new_var_dims) > 0:
            new_var_dims.insert(0, time_dim_name)
        return new_var_dims

    def _normalize_var_infos(self, var_infos: Dict[str, Dict[str, Any]]) -> \
            Dict[str, Dict[str, Any]]:
        if self._normalize_data:
            return normalize_var_infos(var_infos.copy())
        return var_infos

    def _normalize_coord_names(self, coord_names: List[str]):
        if self._normalize_data:
            return normalize_coord_names(coord_names.copy())
        return coord_names


class CciOdpDatasetOpener(CciOdpDataOpener):

    def __init__(self, normalize_data: bool = True, **cdc_params):
        super().__init__(
            CciOdp(**cdc_params, data_type='dataset'),
            DATASET_OPENER_ID,
            DATASET_TYPE,
            normalize_data=normalize_data,
        )

    @classmethod
    def get_data_types(cls) -> Tuple[DataType, ...]:
        return DATASET_TYPE,

    def _get_data_descriptor_from_metadata(self,
                                           data_id: str,
                                           metadata: dict) -> DatasetDescriptor:
        ds_metadata = metadata.copy()
        is_climatology = \
            ds_metadata.get('time_frequency', '') == 'climatology' and \
            'AEROSOL' in data_id
        dims = self._normalize_dims(ds_metadata.get('dimensions', {}))
        bounds_dim_name = None
        for dim_name, dim_size in dims.items():
            if dim_size == 2:
                bounds_dim_name = dim_name
                break
        if not bounds_dim_name and not is_climatology:
            bounds_dim_name = 'bnds'
            dims['bnds'] = 2
        temporal_resolution = get_temporal_resolution_from_id(data_id)
        dataset_info = self._cci_cdc.get_dataset_info(data_id, ds_metadata)
        spatial_resolution = dataset_info['y_res']
        if spatial_resolution <= 0:
            spatial_resolution = None
        bbox = dataset_info['bbox']
        crs = dataset_info['crs']
        # only use date parts of times
        if is_climatology:
            temporal_coverage = None
        else:
            temporal_coverage = (
                dataset_info['temporal_coverage_start'].split('T')[0]
                if dataset_info['temporal_coverage_start'] else None,
                dataset_info['temporal_coverage_end'].split('T')[0]
                if dataset_info['temporal_coverage_end'] else None
            )
        var_infos = self._normalize_var_infos(
            ds_metadata.get('variable_infos', {})
        )
        coord_names = self._normalize_coord_names(dataset_info['coord_names'])

        time_dim_name = "time"
        if is_climatology:
            time_dim_name = 'month'
        if "Time" in dims.keys():
            time_dim_name = "Time"
        var_descriptors = self._get_variable_descriptors(
            dataset_info['var_names'], var_infos, time_dim_name
        )
        coord_descriptors = self._get_variable_descriptors(coord_names,
                                                           var_infos,
                                                           time_dim_name,
                                                           normalize_dims=False)
        if time_dim_name not in coord_descriptors.keys() and \
                't' not in coord_descriptors.keys():
            if is_climatology:
                coord_descriptors[time_dim_name] = VariableDescriptor(
                    time_dim_name, dtype='int8', dims=('time_dim_name',)
                )
            else:
                time_attrs = {
                    "units": "seconds since 1970-01-01T00:00:00Z",
                    "calendar": "proleptic_gregorian",
                    "standard_name": "time",
                    "shape": [dims[time_dim_name]],
                    "size": dims[time_dim_name]
                }
                coord_descriptors[time_dim_name] = VariableDescriptor(
                    time_dim_name, dtype='int64', dims=(time_dim_name,),
                    attrs=time_attrs
                )
                if 'time_bnds' in coord_descriptors.keys():
                    coord_descriptors.pop('time_bnds')
                if 'time_bounds' in coord_descriptors.keys():
                    coord_descriptors.pop('time_bounds')
                time_bnds_attrs = {
                    "units": "seconds since 1970-01-01T00:00:00Z",
                    "calendar": "proleptic_gregorian",
                    "standard_name": "time_bnds",
                    "shape": [2, dims[time_dim_name]],
                    "size": 2 * dims[time_dim_name]
                }
                coord_descriptors['time_bnds'] = \
                    VariableDescriptor('time_bnds',
                                       dtype='int64',
                                       dims=(time_dim_name, bounds_dim_name),
                                       attrs=time_bnds_attrs)

        if 'variables' in ds_metadata:
            ds_metadata.pop('variables')
        ds_metadata.pop('dimensions')
        ds_metadata.pop('variable_infos')
        attrs = ds_metadata.get('attributes', {}).get('NC_GLOBAL', {})
        ds_metadata.pop('attributes')
        attrs.update(ds_metadata)
        self._remove_irrelevant_metadata_attributes(attrs)
        descriptor = DatasetDescriptor(data_id,
                                       data_type=self._data_type,
                                       crs=crs,
                                       dims=dims,
                                       coords=coord_descriptors,
                                       data_vars=var_descriptors,
                                       attrs=attrs,
                                       bbox=bbox,
                                       spatial_res=spatial_resolution,
                                       time_range=temporal_coverage,
                                       time_period=temporal_resolution)
        data_schema = self._get_open_data_params_schema(descriptor)
        descriptor.open_params_schema = data_schema
        return descriptor

    def open_data(self, data_id: str, **open_params) -> Any:
        cci_schema = self.get_open_data_params_schema(data_id)
        cci_schema.validate_instance(open_params)
        cube_kwargs, open_params = cci_schema.process_kwargs_subset(
            open_params, ('variable_names',
                          'time_range',
                          'bbox')
        )
        chunk_store = CciChunkStore(self._cci_cdc, data_id, cube_kwargs)
        ds = xr.open_zarr(chunk_store, consolidated=False)
        ds.zarr_store.set(chunk_store)
        ds = self._normalize_dataset(ds)
        return ds

    def _get_open_data_params_schema(
            self, dsd: DatasetDescriptor = None
    ) -> JsonObjectSchema:
        # noinspection PyUnresolvedReferences
        dataset_params = dict(
            normalize_data=JsonBooleanSchema(default=True),
            variable_names=JsonArraySchema(items=JsonStringSchema(
                enum=dsd.data_vars.keys() if dsd and dsd.data_vars else None))
        )
        if dsd:
            min_date = dsd.time_range[0] if dsd.time_range else None
            max_date = dsd.time_range[1] if dsd.time_range else None
            if min_date or max_date:
                dataset_params['time_range'] = \
                    JsonDateSchema.new_range(min_date, max_date)
        else:
            dataset_params['time_range'] = JsonDateSchema.new_range(None, None)
        if dsd:
            try:
                if pyproj.CRS.from_string(dsd.crs).is_geographic:
                    min_lon = dsd.bbox[0] if dsd and dsd.bbox else -180
                    min_lat = dsd.bbox[1] if dsd and dsd.bbox else -90
                    max_lon = dsd.bbox[2] if dsd and dsd.bbox else 180
                    max_lat = dsd.bbox[3] if dsd and dsd.bbox else 90
                    bbox = JsonArraySchema(items=(
                        JsonNumberSchema(minimum=min_lon, maximum=max_lon),
                        JsonNumberSchema(minimum=min_lat, maximum=max_lat),
                        JsonNumberSchema(minimum=min_lon, maximum=max_lon),
                        JsonNumberSchema(minimum=min_lat, maximum=max_lat)))
                    dataset_params['bbox'] = bbox
            except pyproj.exceptions.CRSError:
                # do not set bbox then
                pass
        cci_schema = JsonObjectSchema(
            properties=dict(**dataset_params),
            required=[
            ],
            additional_properties=False
        )
        return cci_schema


class CciOdpDataFrameOpener(CciOdpDataOpener):

    def __init__(self, **cdc_params):
        super().__init__(
            CciOdp(**cdc_params, data_type='geodataframe'),
            DATAFRAME_OPENER_ID,
            GEO_DATA_FRAME_TYPE
        )

    @classmethod
    def get_data_types(cls) -> Tuple[DataType, ...]:
        return GEO_DATA_FRAME_TYPE,

    def _get_data_descriptor_from_metadata(self,
                                           data_id: str,
                                           metadata: dict) -> GeoDataFrameDescriptor:
        ds_metadata = metadata.copy()
        temporal_resolution = get_temporal_resolution_from_id(data_id)
        dataset_info = self._cci_cdc.get_dataset_info(data_id, ds_metadata)
        bbox = dataset_info['bbox']
        crs = dataset_info['crs']
        temporal_coverage = (
            dataset_info['temporal_coverage_start'].split('T')[0]
            if dataset_info['temporal_coverage_start'] else None,
            dataset_info['temporal_coverage_end'].split('T')[0]
            if dataset_info['temporal_coverage_end'] else None
        )
        feature_schema = self._get_feature_schema(ds_metadata)
        descriptor = GeoDataFrameDescriptor(data_id,
                                            data_type=self._data_type,
                                            crs=crs,
                                            bbox=bbox,
                                            time_range=temporal_coverage,
                                            time_period=temporal_resolution,
                                            feature_schema=feature_schema)
        data_schema = self._get_open_data_params_schema(descriptor)
        descriptor.open_params_schema = data_schema
        return descriptor

    def get_title(self, data_id: str) -> str:
        ds_metadata = self._cci_cdc.get_dataset_metadata(data_id)
        return ds_metadata.get("title", data_id)

    @staticmethod
    def _get_feature_schema(metadata: dict):
        int_data_types = ['uint8', 'uint16', 'uint32', 'int8', 'int16', 'int32']
        features = dict()
        for var_name, var_dict in metadata.get("variable_infos", {}).items():
            dt = var_dict.get("datatype")
            if var_name in ["lat", "lon", "latitude", "longitude"]:
                continue
            if dt == "float32" or dt == "float64":
                features[var_name] = JsonNumberSchema()
            elif dt in int_data_types:
                features[var_name] = JsonIntegerSchema()
            else:
                features[var_name] = JsonObjectSchema()
        features["time"] = JsonDateSchema()
        features["geometry"] = JsonStringSchema(pattern="POINT(* *)")
        feature_schema = JsonObjectSchema(
            properties=dict(**features),
            required=[
            ],
            additional_properties=False
        )
        return feature_schema

    def open_data(self, data_id: str, **open_params) -> Any:
        cci_schema = self.get_open_data_params_schema(data_id)
        cci_schema.validate_instance(open_params)
        frame_kwargs, open_params = cci_schema.process_kwargs_subset(
            open_params, ('variable_names',
                          'time_range',
                          'bbox')
        )
        data_frame_accessor = DataFrameAccessor(self._cci_cdc, data_id, frame_kwargs)
        return data_frame_accessor.get_geodataframe()

    def _get_open_data_params_schema(
            self,
            gfd: GeoDataFrameDescriptor = None
    ) -> JsonObjectSchema:
        # noinspection PyUnresolvedReferences
        geodataframe_params = dict()
        if gfd:
            feature_schema = gfd.feature_schema.to_dict()
            var_names = list(feature_schema.get("properties", {}).keys())
            var_names.remove("geometry")
            var_names.remove("time")
            geodataframe_params["variable_names"] = JsonArraySchema(
                items=JsonStringSchema(enum=var_names)
            )
        if gfd:
            min_date = gfd.time_range[0] if gfd.time_range else None
            max_date = gfd.time_range[1] if gfd.time_range else None
            if min_date or max_date:
                geodataframe_params['time_range'] = \
                    JsonDateSchema.new_range(min_date, max_date)
        else:
            geodataframe_params['time_range'] = JsonDateSchema.new_range(None, None)
        if gfd:
            try:
                if pyproj.CRS.from_string(gfd.crs).is_geographic:
                    min_lon = gfd.bbox[0] if gfd and gfd.bbox else -180
                    min_lat = gfd.bbox[1] if gfd and gfd.bbox else -90
                    max_lon = gfd.bbox[2] if gfd and gfd.bbox else 180
                    max_lat = gfd.bbox[3] if gfd and gfd.bbox else 90
                    bbox = JsonArraySchema(items=(
                        JsonNumberSchema(minimum=min_lon, maximum=max_lon),
                        JsonNumberSchema(minimum=min_lat, maximum=max_lat),
                        JsonNumberSchema(minimum=min_lon, maximum=max_lon),
                        JsonNumberSchema(minimum=min_lat, maximum=max_lat))
                    )
                    geodataframe_params['bbox'] = bbox
            except pyproj.exceptions.CRSError:
                # do not set bbox then
                pass
        cci_schema = JsonObjectSchema(
            properties=dict(**geodataframe_params),
            required=[
            ],
            additional_properties=False
        )
        return cci_schema


class CciOdpVectorDataCubeOpener(CciOdpDataOpener):

    def __init__(self, normalize_data: bool = True, **cdc_params):
        super().__init__(
            CciOdp(**cdc_params, data_type='vectordatacube'),
            VECTORDATACUBE_OPENER_ID,
            VECTOR_DATA_CUBE_TYPE,
            normalize_data=normalize_data,
        )

    @classmethod
    def get_data_types(cls) -> Tuple[DataType, ...]:
        return VECTOR_DATA_CUBE_TYPE,

    def _get_data_descriptor_from_metadata(
            self, data_id: str, metadata: dict
    ) -> VectorDataCubeDescriptor:
        ds_metadata = metadata.copy()
        dims = self._normalize_dims(ds_metadata.get('dimensions', {}))
        temporal_resolution = get_temporal_resolution_from_id(data_id)
        dataset_info = self._cci_cdc.get_dataset_info(data_id, ds_metadata)
        bbox = dataset_info['bbox']
        crs = dataset_info['crs']
        temporal_coverage = (
            dataset_info['temporal_coverage_start'].split('T')[0]
            if dataset_info['temporal_coverage_start'] else None,
            dataset_info['temporal_coverage_end'].split('T')[0]
            if dataset_info['temporal_coverage_end'] else None
        )
        var_infos = self._normalize_var_infos(
            ds_metadata.get('variable_infos', {})
        )
        var_names = dataset_info["var_names"]
        lat_lons = [("lat", "lon"), ("latitude", "longitude")]
        for lat_lon in lat_lons:
            if lat_lon[0] in var_names and lat_lon[1] in var_names:
                var_names.remove(lat_lon[0])
                var_names.remove(lat_lon[1])
            break
        coord_names = self._normalize_coord_names(dataset_info['coord_names'])
        time_dim_name = 'time'
        var_descriptors = self._get_variable_descriptors(
            dataset_info['var_names'], var_infos, time_dim_name
        )
        coord_descriptors = self._get_variable_descriptors(coord_names,
                                                           var_infos,
                                                           time_dim_name,
                                                           normalize_dims=False)
        if 'variables' in ds_metadata:
            ds_metadata.pop('variables')
        ds_metadata.pop('dimensions')
        ds_metadata.pop('variable_infos')
        attrs = ds_metadata.get('attributes', {}).get('NC_GLOBAL', {})
        ds_metadata.pop('attributes')
        attrs.update(ds_metadata)
        self._remove_irrelevant_metadata_attributes(attrs)
        descriptor = VectorDataCubeDescriptor(data_id,
                                              data_type=self._data_type,
                                              crs=crs,
                                              dims=dims,
                                              coords=coord_descriptors,
                                              data_vars=var_descriptors,
                                              attrs=attrs,
                                              bbox=bbox,
                                              time_range=temporal_coverage,
                                              time_period=temporal_resolution)
        data_schema = self._get_open_data_params_schema(descriptor)
        descriptor.open_params_schema = data_schema
        return descriptor

    def _get_open_data_params_schema(
            self, vdcd: VectorDataCubeDescriptor = None
    ) -> JsonObjectSchema:
        # noinspection PyUnresolvedReferences
        vectordatacube_params = dict(
            normalize_data=JsonBooleanSchema(default=True),
            variable_names=JsonArraySchema(items=JsonStringSchema(
                enum=vdcd.data_vars.keys() if vdcd and vdcd.data_vars else None))
        )
        # TODO add this when spatial and temporal subsetting are supported
        # if vdcd:
        #     min_date = vdcd.time_range[0] if vdcd.time_range else None
        #     max_date = vdcd.time_range[1] if vdcd.time_range else None
        #     if min_date or max_date:
        #         vectordatacube_params['time_range'] = \
        #             JsonDateSchema.new_range(min_date, max_date)
        # else:
        #     vectordatacube_params['time_range'] = JsonDateSchema.new_range(None, None)
        # if vdcd:
        #     try:
        #         if pyproj.CRS.from_string(vdcd.crs).is_geographic:
        #             min_lon = vdcd.bbox[0] if vdcd and vdcd.bbox else -180
        #             min_lat = vdcd.bbox[1] if vdcd and vdcd.bbox else -90
        #             max_lon = vdcd.bbox[2] if vdcd and vdcd.bbox else 180
        #             max_lat = vdcd.bbox[3] if vdcd and vdcd.bbox else 90
        #             bbox = JsonArraySchema(items=(
        #                 JsonNumberSchema(minimum=min_lon, maximum=max_lon),
        #                 JsonNumberSchema(minimum=min_lat, maximum=max_lat),
        #                 JsonNumberSchema(minimum=min_lon, maximum=max_lon),
        #                 JsonNumberSchema(minimum=min_lat, maximum=max_lat)))
        #             vectordatacube_params['bbox'] = bbox
        #     except pyproj.exceptions.CRSError:
                # do not set bbox then
                # pass
        cci_schema = JsonObjectSchema(
            properties=dict(**vectordatacube_params),
            required=[
            ],
            additional_properties=False
        )
        return cci_schema

    def open_data(self, data_id: str, **open_params) -> Any:
        cci_schema = self.get_open_data_params_schema(data_id)
        cci_schema.validate_instance(open_params)
        cube_kwargs, open_params = cci_schema.process_kwargs_subset(
            open_params, ('variable_names',
                          'time_range',
                          'bbox')
        )
        chunk_store = CciChunkStore(self._cci_cdc, data_id, cube_kwargs)
        ds = xr.open_zarr(chunk_store, consolidated=False)

        def _convert_to_point(chunk):
            return [Point(point_dict.get("coordinates")) for point_dict in chunk]

        da = xr.apply_ufunc(_convert_to_point, ds.geometry,
                            dask='parallelized', output_dtypes=["object"])
        ds = ds.assign_coords(geometry=da)
        ds = ds.set_xindex("geometry", xvec.GeometryIndex)
        ds.zarr_store.set(chunk_store)
        return ds


class CciOdpDataStore(DataStore):

    def __init__(self, normalize_data=True, **store_params):
        cci_schema = self.get_data_store_params_schema()
        cci_schema.validate_instance(store_params)
        cdc_kwargs, store_params = cci_schema.process_kwargs_subset(
            store_params, ('endpoint_url', 'endpoint_description_url',
                           'enable_warnings', 'num_retries',
                           'retry_backoff_max', 'retry_backoff_base',
                           'user_agent')
        )
        dataset_opener = CciOdpDatasetOpener(
            normalize_data=normalize_data,
            **cdc_kwargs
        )
        dataframe_opener = CciOdpDataFrameOpener(
            **cdc_kwargs
        )
        vectordatacube_opener = CciOdpVectorDataCubeOpener(
            **cdc_kwargs
        )
        self._openers = {
            DATASET_OPENER_ID: dataset_opener,
            DATAFRAME_OPENER_ID: dataframe_opener,
            VECTORDATACUBE_OPENER_ID: vectordatacube_opener
        }
        dataset_states_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'data/dataset_states.json'
        )
        with open(dataset_states_file, 'r') as fp:
            self._dataset_states = json.load(fp)

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        cciodp_params = dict(
            endpoint_url=JsonStringSchema(default=OPENSEARCH_CEDA_URL),
            endpoint_description_url=JsonStringSchema(default=CCI_ODD_URL),
            enable_warnings=JsonBooleanSchema(
                default=False, title='Whether to output warnings'
            ),
            num_retries=JsonIntegerSchema(
                default=DEFAULT_NUM_RETRIES, minimum=0,
                title='Number of retries when requesting data fails'
            ),
            retry_backoff_max=JsonIntegerSchema(
                default=DEFAULT_RETRY_BACKOFF_MAX, minimum=0
            ),
            retry_backoff_base=JsonNumberSchema(
                default=DEFAULT_RETRY_BACKOFF_BASE, exclusive_minimum=1.0
            ),
            user_agent=JsonStringSchema(default=None)
        )
        return JsonObjectSchema(
            properties=dict(**cciodp_params),
            required=None,
            additional_properties=False
        )

    @classmethod
    def get_data_types(cls) -> Tuple[str, ...]:
        return (DATASET_TYPE.alias, GEO_DATA_FRAME_TYPE.alias,
                VECTOR_DATA_CUBE_TYPE.alias)

    def get_data_types_for_data(self, data_id: str) -> Tuple[str, ...]:
        if self.has_data(data_id, data_type=DATASET_TYPE.alias):
            return DATASET_TYPE.alias,
        if self.has_data(data_id, data_type=GEO_DATA_FRAME_TYPE.alias):
            return GEO_DATA_FRAME_TYPE.alias,
        if self.has_data(data_id, data_type=VECTOR_DATA_CUBE_TYPE.alias):
            return VECTOR_DATA_CUBE_TYPE.alias,
        raise DataStoreError(
            f'Data resource {data_id!r} does not exist in store'
        )

    def _get_openers(
            self, opener_id: str = None, data_type: str = None
    ) -> List[CciOdpDataOpener]:
        self._assert_valid_opener_id(opener_id)
        try:
            self._assert_valid_data_type(data_type)
        except DataStoreError:
            return []
        if opener_id is not None:
            opener = self._openers[opener_id]
            if data_type is not None:
                opener_types = opener.get_data_types()
                for opener_type in opener_types:
                    if opener_type.is_super_type_of(data_type):
                        return [opener]
                raise ValueError(f"Opener '{opener_id}' is not compatible with "
                                 f"data type '{data_type}'.")
        if data_type is not None:
            if data_type == "vectordatacube":
                return [self._openers[VECTORDATACUBE_OPENER_ID]]
            if DATASET_TYPE.is_super_type_of(data_type):
                return [self._openers[DATASET_OPENER_ID]]
            if GEO_DATA_FRAME_TYPE.is_super_type_of(data_type):
                return [self._openers[DATAFRAME_OPENER_ID]]
        return list(self._openers.values())

    def get_data_ids(self,
                     data_type: str = None,
                     include_attrs: Container[str] | bool = False,) -> \
            Union[Iterator[str], Iterator[Tuple[str, Dict[str, Any]]]]:
        openers = self._get_openers(data_type=data_type)
        data_ids = []
        for opener in openers:
            data_ids.extend(opener.dataset_names)
        for data_id in data_ids:
            if include_attrs is None or not include_attrs:
                yield data_id
            else:
                attrs = {}
                if isinstance(include_attrs, Container):
                    for attr in include_attrs:
                        value = self._dataset_states.get(data_id, {}).get(attr)
                        if value is not None:
                            attrs[attr] = value
                elif include_attrs:
                    attrs = self._dataset_states.get(data_id, {})
                yield data_id, attrs

    def has_data(self, data_id: str, data_type: str = None) -> bool:
        openers = self._get_openers(data_type=data_type)
        for opener in openers:
            if opener.has_data(data_id):
                return True
        return False

    def describe_data(
            self, data_id: str, data_type: str = None
    ) -> DataDescriptor:
        openers = self._get_openers(data_type=data_type)
        for opener in openers:
            if opener.has_data(data_id):
                return opener.describe_data(data_id)
        raise ValueError(f"No opener provides data with id {data_id}")

    @classmethod
    def get_search_params_schema(
            cls, data_type: str = None
    ) -> JsonObjectSchema:
        cls._assert_valid_data_type(data_type)
        if data_type is not None:
            data_ids = CciOdp(data_type=data_type).dataset_names
        else:
            data_ids = CciOdp(data_type=DATASET_TYPE.alias).dataset_names + \
                       CciOdp(data_type=GEO_DATA_FRAME_TYPE.alias).dataset_names + \
                       CciOdp(data_type=VECTOR_DATA_CUBE_TYPE.alias).dataset_names
        ecvs = set([data_id.split('.')[1] for data_id in data_ids])
        frequencies = set(
            [data_id.split('.')[2].replace('-days', ' days').
                 replace('mon', 'month').replace('-yrs', ' years').
                 replace('yr', 'year') for data_id in data_ids]
        )
        processing_levels = set([data_id.split('.')[3] for data_id in data_ids])
        data_types = set([data_id.split('.')[4] for data_id in data_ids])
        sensors = set([data_id.split('.')[5] for data_id in data_ids])
        platforms = set([data_id.split('.')[6] for data_id in data_ids])
        product_strings = set([data_id.split('.')[7] for data_id in data_ids])
        product_versions = set([data_id.split('.')[8].replace('-', '.')
                                for data_id in data_ids])
        search_params = dict(
            start_date=JsonStringSchema(format='date-time'),
            end_date=JsonStringSchema(format='date-time'),
            bbox=JsonArraySchema(items=(JsonNumberSchema(),
                                        JsonNumberSchema(),
                                        JsonNumberSchema(),
                                        JsonNumberSchema())),
            cci_attrs=JsonObjectSchema(
                properties=dict(
                    ecv=JsonStringSchema(enum=ecvs),
                    frequency=JsonStringSchema(enum=frequencies),
                    institute=JsonStringSchema(enum=_INSTITUTES),
                    processing_level=JsonStringSchema(enum=processing_levels),
                    product_string=JsonStringSchema(enum=product_strings),
                    product_version=JsonStringSchema(enum=product_versions),
                    data_type=JsonStringSchema(enum=data_types),
                    sensor=JsonStringSchema(enum=sensors),
                    platform=JsonStringSchema(enum=platforms)
                ),
                additional_properties=False
            )
        )
        search_schema = JsonObjectSchema(
            properties=dict(**search_params),
            additional_properties=False)
        return search_schema

    def search_data(
            self, data_type: str = None, **search_params
    ) -> Iterator[DatasetDescriptor]:
        search_schema = self.get_search_params_schema()
        search_schema.validate_instance(search_params)
        openers = self._get_openers(data_type=data_type)
        desc_iterators = []
        for opener in openers:
            desc_iterators.extend(opener.search_data(**search_params))
        return desc_iterators

    def get_data_opener_ids(
            self, data_id: str = None, data_type: str = None,
    ) -> Tuple[str, ...]:
        self._assert_valid_data_type(data_type)
        if data_id is not None \
                and not self.has_data(data_id, data_type=data_type):
            raise DataStoreError(f'Data resource {data_id!r}'
                                 f' is not available.')
        if data_type is None:
            return list(self._openers.keys())
        opener_ids = []
        for opener_id, opener in self._openers.items():
            for opener_data_type in opener.get_data_types():
                if opener_data_type.is_super_type_of(data_type):
                    opener_ids.append(opener_id)
                    break
        if len(opener_ids) == 0:
            raise DataStoreError(f'Data resource {data_id!r}'
                                 f' is not available as type {data_type!r}.')
        return tuple(opener_ids)

    def get_open_data_params_schema(
            self, data_id: str = None, opener_id: str = None
    ) -> JsonObjectSchema:
        openers = self._get_openers(opener_id=opener_id)
        if data_id is not None:
            for opener in openers:
                if opener.has_data(data_id):
                    return opener.get_open_data_params_schema(data_id)
            raise ValueError(f"No schema for '{data_id}' with provided opener ids.")
        # TODO find solution for when neither data id nor opener id have been provided
        # Have less specific schema?
        return openers[0].get_open_data_params_schema(data_id)

    def open_data(
            self, data_id: str, opener_id: str = None, **open_params
    ) -> xr.Dataset:
        openers = self._get_openers(opener_id=opener_id)
        for opener in openers:
            if opener.has_data(data_id):
                return opener.open_data(data_id, **open_params)
        raise ValueError(f"Could not open data '{data_id}'. No opener found.")

    ############################################################################
    # Implementation helpers

    @classmethod
    def _is_valid_data_type(cls, data_type: str) -> bool:
        return data_type is None \
               or data_type == "vectordatacube" \
               or DATASET_TYPE.is_super_type_of(data_type) \
               or GEO_DATA_FRAME_TYPE.is_super_type_of(data_type)

    @classmethod
    def _assert_valid_data_type(cls, data_type):
        if not cls._is_valid_data_type(data_type):
            raise DataStoreError(
                f'Data type must be {DATASET_TYPE!r}, {GEO_DATA_FRAME_TYPE!r}, '
                f'or {VECTOR_DATA_CUBE_TYPE!r}, but got {data_type!r}')

    def _assert_valid_opener_id(self, opener_id):
        if opener_id is not None and opener_id not in list(self._openers.keys()):
            raise DataStoreError(
                f'Data opener identifier must be {DATASET_OPENER_ID!r}, '
                f'{DATAFRAME_OPENER_ID!r}, or {VECTORDATACUBE_OPENER_ID!r},'
                f'but got {opener_id!r}'
            )



import numpy as np
import os
import pandas as pd
import unittest
from unittest import skip, skipIf

from xcube_cci.cciodp import get_res
from xcube_cci.cciodp import CciOdp
from xcube_cci.cciodp import find_datetime_format
from xcube_cci.constants import OPENSEARCH_CEDA_URL

AEROSOL_ID = 'esacci.AEROSOL.day.L3C.AER_PRODUCTS.AATSR.Envisat.ORAC.04-01-.r1'
CLOUD_ID = 'esacci.CLOUD.mon.L3C.CLD_PRODUCTS.MODIS.Terra.MODIS_TERRA.2-0.r1'
OC_DAY_ID = 'esacci.OC.day.L3S.CHLOR_A.multi-sensor.multi-platform.MERGED.' \
           '6-0.geographic'
OC_DAY_SIN_ID = 'esacci.OC.day.L3S.K_490.multi-sensor.multi-platform.' \
                'MERGED.6-0.sinusoidal'
OC_DAY_31_ID = 'esacci.OC.day.L3S.CHLOR_A.multi-sensor.multi-platform.' \
               'MERGED.3-1.geographic'
OC_MON_ID = \
    'esacci.OC.day.L3S.K_490.multi-sensor.multi-platform.MERGED.6-0.sinusoidal'
OC_MON_31_ID = 'esacci.OC.mon.L3S.CHLOR_A.multi-sensor.multi-platform.MERGED.' \
               '3-1.geographic'
OC_5DAYS_GEO_ID = 'esacci.OC.5-days.L3S.K_490.multi-sensor.multi-platform.' \
              'MERGED.6-0.geographic'
OC_5DAYS_SIN_ID = 'esacci.OC.5-days.L3S.K_490.multi-sensor.multi-platform.' \
                  'MERGED.6-0.sinusoidal'
OC_5DAYS_CHLOR_SIN_ID = 'esacci.OC.5-days.L3S.CHLOR_A.multi-sensor.' \
                        'multi-platform.MERGED.6-0.geographic'
OC_5DAYS_CHLOR_SIN_31_ID = 'esacci.OC.5-days.L3S.CHLOR_A.multi-sensor.' \
                           'multi-platform.MERGED.3-1.geographic'
OC_8DAYS_CHLOR_31_ID = 'esacci.OC.8-days.L3S.CHLOR_A.multi-sensor.' \
                       'multi-platform.MERGED.3-1.geographic'
OZONE_ID = 'esacci.OZONE.mon.L3.NP.multi-sensor.multi-platform.MERGED.fv0002.r1'
SEAICE_ID = 'esacci.SEAICE.mon.L3C.SITHICK.SIRAL.CryoSat-2.NH25KMEASE2.2-0.r1'
SST_ID = 'esacci.SST.satellite-orbit-frequency.L3U.SSTskin.AVHRR-3.NOAA-19.' \
         'AVHRR19_G.2-1.r1'


class CciCdcTest(unittest.TestCase):

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_data_chunk(self):
        cci_cdc = CciOdp()
        ds_id = cci_cdc.get_dataset_id(OZONE_ID)
        request = dict(parentIdentifier=ds_id,
                       startDate='1997-05-01T00:00:00',
                       endDate='1997-05-02T00:00:00',
                       varNames=['surface_pressure'],
                       drsId=OZONE_ID
                       )
        dim_indexes = (slice(None, None), slice(0, 179), slice(0, 359))
        data = cci_cdc.get_data_chunk(request, dim_indexes)
        self.assertIsNotNone(data)
        data_array = np.frombuffer(data, dtype=np.float32)
        self.assertEqual(64261, len(data_array))
        self.assertAlmostEqual(1024.4185, data_array[-1], 4)

        # check whether data type has been converted to
        request = dict(parentIdentifier=ds_id,
                       startDate='1997-05-01T00:00:00',
                       endDate='1997-05-02T00:00:00',
                       varNames=['layers'],
                       drsId=OZONE_ID)
        dim_indexes = (slice(None, None), slice(0, 179), slice(0, 359))
        data = cci_cdc.get_data_chunk(request, dim_indexes)
        self.assertIsNotNone(data)
        data_array = np.frombuffer(data, dtype=np.int64)
        self.assertEqual(16, len(data_array))
        self.assertAlmostEqual(15, data_array[-2])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_dataset_names(self):
        cci_cdc = CciOdp()
        dataset_names = cci_cdc.dataset_names
        self.assertIsNotNone(dataset_names)
        list(dataset_names)
        self.assertTrue(len(dataset_names) > 250)
        self.assertTrue(AEROSOL_ID in dataset_names)
        self.assertTrue(OC_MON_ID in dataset_names)
        self.assertTrue(SST_ID in dataset_names)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_var_and_coord_names(self):
        cci_cdc = CciOdp()
        var_names, coord_names = cci_cdc.var_and_coord_names(OC_5DAYS_GEO_ID)
        self.assertIsNotNone(var_names)
        self.assertEqual(
            ['MERIS_nobs_sum', 'MODISA_nobs_sum', 'OLCI-A_nobs_sum',
             'OLCI-B_nobs_sum', 'SeaWiFS_nobs_sum', 'VIIRS_nobs_sum', 'crs',
             'kd_490', 'kd_490_bias', 'kd_490_rmsd', 'total_nobs_sum'],
            var_names
        )
        self.assertIsNotNone(coord_names)
        self.assertEqual(['lat', 'lon', 'time'], coord_names)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_dataset_info(self):
        cci_cdc = CciOdp()
        dataset_info = cci_cdc.get_dataset_info(CLOUD_ID)
        self.assertIsNotNone(dataset_info)
        self.assertTrue('y_res' in dataset_info)
        self.assertEqual(0.5, dataset_info['y_res'])
        self.assertTrue('x_res' in dataset_info)
        self.assertEqual(0.5, dataset_info['x_res'])
        self.assertTrue('bbox' in dataset_info)
        self.assertEqual((-180.0, -90.0, 180.0, 90.0), dataset_info['bbox'])
        self.assertTrue('var_names' in dataset_info)
        self.assertEqual([
            'nobs', 'nobs_day', 'nobs_clear_day', 'nobs_cloudy_day',
            'nobs_clear_night', 'nobs_cloudy_night', 'nobs_clear_twl',
            'nobs_cloudy_twl', 'nobs_cloudy', 'nretr_cloudy',
            'nretr_cloudy_liq', 'nretr_cloudy_ice', 'nretr_cloudy_day',
            'nretr_cloudy_day_liq', 'nretr_cloudy_day_ice', 'nretr_cloudy_low',
            'nretr_cloudy_mid', 'nretr_cloudy_high', 'cfc', 'cfc_std',
            'cfc_prop_unc', 'cfc_corr_unc', 'cfc_unc', 'cfc_low', 'cfc_mid',
            'cfc_high', 'cfc_day', 'cfc_night', 'cfc_twl', 'ctt', 'ctt_std',
            'ctt_prop_unc', 'ctt_corr_unc', 'ctt_unc', 'ctt_corrected',
            'ctt_corrected_std', 'ctt_corrected_prop_unc',
            'ctt_corrected_corr_unc', 'ctt_corrected_unc', 'stemp', 'stemp_std',
            'stemp_prop_unc', 'stemp_corr_unc', 'stemp_unc', 'cth', 'cth_std',
            'cth_prop_unc', 'cth_corr_unc', 'cth_unc', 'cth_corrected',
            'cth_corrected_std', 'cth_corrected_prop_unc',
            'cth_corrected_corr_unc', 'cth_corrected_unc', 'ctp', 'ctp_std',
            'ctp_prop_unc', 'ctp_corr_unc', 'ctp_unc', 'ctp_log',
            'ctp_corrected', 'ctp_corrected_std', 'ctp_corrected_prop_unc',
            'ctp_corrected_corr_unc', 'ctp_corrected_unc', 'cph', 'cph_std',
            'cph_day', 'cph_day_std', 'cer', 'cer_std', 'cer_prop_unc',
            'cer_corr_unc', 'cer_unc', 'cot', 'cot_log', 'cot_std',
            'cot_prop_unc', 'cot_corr_unc', 'cot_unc', 'cee', 'cee_std',
            'cee_prop_unc', 'cee_corr_unc', 'cee_unc', 'cla_vis006',
            'cla_vis006_std', 'cla_vis006_prop_unc', 'cla_vis006_corr_unc',
            'cla_vis006_unc', 'cla_vis006_liq', 'cla_vis006_liq_std',
            'cla_vis006_liq_unc', 'cla_vis006_ice', 'cla_vis006_ice_std',
            'cla_vis006_ice_unc', 'cla_vis008', 'cla_vis008_std',
            'cla_vis008_prop_unc', 'cla_vis008_corr_unc', 'cla_vis008_unc',
            'cla_vis008_liq', 'cla_vis008_liq_std', 'cla_vis008_liq_unc',
            'cla_vis008_ice', 'cla_vis008_ice_std', 'cla_vis008_ice_unc', 'lwp',
            'lwp_std', 'lwp_prop_unc', 'lwp_corr_unc', 'lwp_unc', 'lwp_allsky',
            'iwp', 'iwp_std', 'iwp_prop_unc', 'iwp_corr_unc', 'iwp_unc',
            'iwp_allsky', 'cer_liq', 'cer_liq_std', 'cer_liq_prop_unc',
            'cer_liq_corr_unc', 'cer_liq_unc', 'cer_ice', 'cer_ice_std',
            'cer_ice_prop_unc', 'cer_ice_corr_unc', 'cer_ice_unc', 'cot_liq',
            'cot_liq_std', 'cot_liq_prop_unc', 'cot_liq_corr_unc',
            'cot_liq_unc', 'cot_ice', 'cot_ice_std', 'cot_ice_prop_unc',
            'cot_ice_corr_unc', 'cot_ice_unc', 'hist2d_cot_ctp', 'hist1d_cot',
            'hist1d_ctp', 'hist1d_ctt', 'hist1d_cer', 'hist1d_cwp',
            'hist1d_cla_vis006', 'hist1d_cla_vis008'],
            dataset_info['var_names']
        )
        self.assertTrue('temporal_coverage_start' in dataset_info)
        self.assertEqual('2000-02-01T00:00:00',
                         dataset_info['temporal_coverage_start'])
        self.assertTrue('temporal_coverage_end' in dataset_info)
        self.assertEqual('2014-12-31T23:59:59',
                         dataset_info['temporal_coverage_end'])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_fetch_data_source_list_json(self):
        cci_cdc = CciOdp()

        async def fetch_data_source_list_json(session):
            return await cci_cdc._fetch_data_source_list_json(
                session, OPENSEARCH_CEDA_URL, {
                    'parentIdentifier': 'cci',
                    'drsId': OC_5DAYS_SIN_ID}
            )

        data_source_list = cci_cdc._run_with_session(
            fetch_data_source_list_json
        )
        self.assertIsNotNone(data_source_list)
        self.assertEqual(1, len(data_source_list))
        self.assertTrue(
            '3bdb21a4cd004e5f8cc148fea5f1d4e3' in data_source_list
        )
        self.assertTrue(len(
            data_source_list['3bdb21a4cd004e5f8cc148fea5f1d4e3'].items()
        ) >= 6)
        self.assertEqual(
            '3bdb21a4cd004e5f8cc148fea5f1d4e3',
            data_source_list['3bdb21a4cd004e5f8cc148fea5f1d4e3'].get(
                'uuid', ''
            )
        )
        self.assertEqual(
            'ESA Ocean Colour Climate Change Initiative (Ocean_Colour_cci): '
            'Global attenuation coefficient for downwelling irradiance (Kd490) '
            'gridded on a sinusoidal projection at 4km resolution, Version 6.0',
            data_source_list['3bdb21a4cd004e5f8cc148fea5f1d4e3'].get(
                'title', ''
            )
        )
        self.assertTrue(
            'variables' in data_source_list['3bdb21a4cd004e5f8cc148fea5f1d4e3']
        )
        self.assertTrue(
            'odd_url' in data_source_list['3bdb21a4cd004e5f8cc148fea5f1d4e3']
        )
        self.assertTrue(
            data_source_list['3bdb21a4cd004e5f8cc148fea5f1d4e3'].
                get('metadata_url', '').endswith(
                '3bdb21a4cd004e5f8cc148fea5f1d4e3.xml'
            )
        )
        self.assertEqual(
            'https://catalogue.ceda.ac.uk/uuid/'
            '3bdb21a4cd004e5f8cc148fea5f1d4e3',
            data_source_list['3bdb21a4cd004e5f8cc148fea5f1d4e3'].get(
                'catalog_url', ''
            )
        )

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_datasets_metadata(self):
        cci_cdc = CciOdp()
        datasets = [OC_DAY_ID, OC_5DAYS_CHLOR_SIN_ID, SEAICE_ID]
        datasets_metadata = cci_cdc.get_datasets_metadata(datasets)
        self.assertIsNotNone(datasets_metadata)
        self.assertEqual(3, len(datasets_metadata))
        self.assertTrue('variables' in datasets_metadata[0])
        self.assertTrue('variable_infos' in datasets_metadata[2])
        self.assertTrue('freeboard' in datasets_metadata[2].
                        get('variable_infos'))
        self.assertEqual(
            {'coordinates': 'time lon lat',
             'grid_mapping': 'Lambert_Azimuthal_Grid',
             'long_name': 'elevation of retracked point above instantaneous '
                          'sea surface height (with snow range corrections)',
             'standard_name': 'sea_ice_freeboard',
             'units': 'm',
             'chunk_sizes': [1, 432, 432],
             'file_chunk_sizes': [1, 432, 432],
             'data_type': 'float32',
             'orig_data_type': 'float32',
             'dimensions': ['time', 'yc', 'xc'],
             'file_dimensions': ['time', 'yc', 'xc'],
             'fill_value': np.nan,
             'size': 8957952,
             'shape': [48, 432, 432]},
            datasets_metadata[2].get('variable_infos').get('freeboard'))
        self.assertEqual('uint16', datasets_metadata[2].get('variable_infos').
                         get('status_flag').get('data_type'))

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    @skip('Disabled while old archive is set up')
    def test_get_drs_metadata(self):
        cci_cdc = CciOdp()
        fid = '12d6f4bdabe144d7836b0807e65aa0e2'
        metadata = {}

        async def set_drs_metadata(session):
            return await cci_cdc._set_drs_metadata(session, fid, metadata)

        cci_cdc._run_with_session(set_drs_metadata)
        variables_dict = metadata['variables']
        drs_uuids = metadata['uuids']
        self.assertIsNotNone(variables_dict)
        self.assertEqual(4, len(variables_dict.items()))
        self.assertTrue(OC_DAY_31_ID in variables_dict)
        self.assertEqual(12, len(variables_dict[OC_DAY_31_ID]))
        variable_names = [
            variable['name'] for variable in variables_dict[OC_DAY_31_ID]
        ]
        self.assertEqual(
            ['chlor_a_log10_bias', 'chlor_a', 'MERIS_nobs', 'MODISA_nobs',
             'SeaWiFS_nobs', 'VIIRS_nobs', 'total_nobs', 'chlor_a_log10_rmsd',
             'lat', 'lon', 'time', 'crs'],
            variable_names)
        self.assertTrue(OC_5DAYS_CHLOR_SIN_31_ID in variables_dict)
        variable_names = [variable['name'] for variable
                          in variables_dict[OC_5DAYS_CHLOR_SIN_31_ID]]
        self.assertEqual(
            ['chlor_a_log10_bias', 'chlor_a', 'MERIS_nobs_sum',
             'MODISA_nobs_sum', 'SeaWiFS_nobs_sum', 'VIIRS_nobs_sum',
             'total_nobs_sum', 'chlor_a_log10_rmsd', 'lat', 'lon', 'time',
             'crs'], variable_names)
        self.assertEqual(4, len(drs_uuids.items()))
        self.assertTrue(OC_DAY_31_ID in drs_uuids)
        self.assertEqual('f13668e057c736410676fcf51983ca9d018108c3',
                         drs_uuids[OC_DAY_31_ID])
        self.assertTrue(OC_5DAYS_CHLOR_SIN_31_ID in drs_uuids)
        self.assertEqual('d43b5fa4c4e0d7edc89f794798aa07824671105b',
                         drs_uuids[OC_5DAYS_CHLOR_SIN_31_ID])
        self.assertTrue(OC_8DAYS_CHLOR_31_ID in drs_uuids)
        self.assertEqual('5082ae00d82255c2389a74593b71ebe686fc1314',
                         drs_uuids[OC_8DAYS_CHLOR_31_ID])
        self.assertTrue(OC_MON_31_ID in drs_uuids)
        self.assertEqual('e7f59d6547d610fdd04ae05cca77ddfee15e3a5a',
                         drs_uuids[OC_MON_31_ID])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_opendap_dataset(self):
        opendap_url = 'http://data.cci.ceda.ac.uk/thredds/dodsC/esacci/' \
                      'aerosol/data/AATSR_SU/L3/v4.21/DAILY/2002/07/' \
                      '20020724-ESACCI-L3C_AEROSOL-AER_PRODUCTS-AATSR_' \
                      'ENVISAT-SU_DAILY-v4.21.nc'
        cci_cdc = CciOdp()
        dataset = cci_cdc.get_opendap_dataset(opendap_url)
        self.assertIsNotNone(dataset)
        self.assertEqual(53, len(list(dataset.keys())))
        self.assertTrue('AOD550_mean' in dataset.keys())
        self.assertEqual('atmosphere_optical_thickness_due_to_ambient_aerosol',
                         dataset['AOD550_mean'].attributes['standard_name'])
        self.assertEqual(('latitude', 'longitude'),
                         dataset['AOD550_mean'].dimensions)
        self.assertEqual(64800, dataset['AOD550_mean'].size)

    def test_get_res(self):
        nc_attrs = dict(geospatial_lat_resolution=24.2,
                        geospatial_lon_resolution=30.1)
        self.assertEqual(24.2, get_res(nc_attrs, 'lat'))
        self.assertEqual(30.1, get_res(nc_attrs, 'lon'))

        nc_attrs = dict(resolution=5.0)
        self.assertEqual(5.0, get_res(nc_attrs, 'lat'))
        self.assertEqual(5.0, get_res(nc_attrs, 'lon'))

        nc_attrs = dict(resolution='12x34 degree')
        self.assertEqual(12.0, get_res(nc_attrs, 'lat'))
        self.assertEqual(34.0, get_res(nc_attrs, 'lon'))

        nc_attrs = dict(spatial_resolution='926.62543305 m')
        self.assertEqual(926.62543305, get_res(nc_attrs, 'lat'))
        self.assertEqual(926.62543305, get_res(nc_attrs, 'lon'))

        nc_attrs = dict(spatial_resolution='60km x 30km at nadir '
                                           '(along-track x across-track)')
        self.assertEqual(60.0, get_res(nc_attrs, 'lat'))
        self.assertEqual(30.0, get_res(nc_attrs, 'lon'))

    def test_find_datetime_format(self):
        time_format, start, end, timedelta = \
            find_datetime_format('fetgzrs2015ydhfbgv')
        self.assertEqual('%Y', time_format)
        self.assertEqual(7, start)
        self.assertEqual(11, end)
        self.assertEqual(1, timedelta.years)
        self.assertEqual(0, timedelta.months)
        self.assertEqual(0, timedelta.days)
        self.assertEqual(0, timedelta.hours)
        self.assertEqual(0, timedelta.minutes)
        self.assertEqual(-1, timedelta.seconds)

        time_format, start, end, timedelta = \
            find_datetime_format('fetz23gxgs20150213ydh391fbgv')
        self.assertEqual('%Y%m%d', time_format)
        self.assertEqual(10, start)
        self.assertEqual(18, end)
        self.assertEqual(0, timedelta.years)
        self.assertEqual(0, timedelta.months)
        self.assertEqual(1, timedelta.days)
        self.assertEqual(0, timedelta.hours)
        self.assertEqual(0, timedelta.minutes)
        self.assertEqual(-1, timedelta.seconds)

        time_format, start, end, timedelta = \
            find_datetime_format('f23gxgs19961130191846y391fbgv')
        self.assertEqual('%Y%m%d%H%M%S', time_format)
        self.assertEqual(7, start)
        self.assertEqual(21, end)
        self.assertEqual(0, timedelta.years)
        self.assertEqual(0, timedelta.months)
        self.assertEqual(0, timedelta.days)
        self.assertEqual(0, timedelta.hours)
        self.assertEqual(0, timedelta.minutes)
        self.assertEqual(0, timedelta.seconds)

        time_format, start, end, timedelta = \
            find_datetime_format('f23gxgdrtgys1983-11-30y391fbgv')
        self.assertEqual('%Y-%m-%d', time_format)
        self.assertEqual(12, start)
        self.assertEqual(22, end)
        self.assertEqual(0, timedelta.years)
        self.assertEqual(0, timedelta.months)
        self.assertEqual(1, timedelta.days)
        self.assertEqual(0, timedelta.hours)
        self.assertEqual(0, timedelta.minutes)
        self.assertEqual(-1, timedelta.seconds)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_variable_data(self):
        cci_cdc = CciOdp()
        dimension_data = cci_cdc.get_variable_data(
            AEROSOL_ID,
            {'latitude': 180,
             'longitude': 360,
             'view': 2,
             'aerosol_type': 10},
            '2002-08-01T00:00:00',
            '2002-08-02T00:00:00')
        self.assertIsNotNone(dimension_data)
        self.assertEqual(dimension_data['latitude']['size'], 180)
        self.assertEqual(dimension_data['latitude']['chunkSize'], 180)
        self.assertEqual(dimension_data['latitude']['data'][0], -89.5)
        self.assertEqual(dimension_data['latitude']['data'][-1], 89.5)
        self.assertEqual(dimension_data['longitude']['size'], 360)
        self.assertEqual(dimension_data['longitude']['chunkSize'], 360)
        self.assertEqual(dimension_data['longitude']['data'][0], -179.5)
        self.assertEqual(dimension_data['longitude']['data'][-1], 179.5)
        self.assertEqual(dimension_data['view']['size'], 2)
        self.assertEqual(dimension_data['view']['chunkSize'], 2)
        self.assertEqual(dimension_data['view']['data'][0], 0)
        self.assertEqual(dimension_data['view']['data'][-1], 1)
        self.assertEqual(dimension_data['aerosol_type']['size'], 10)
        self.assertEqual(dimension_data['aerosol_type']['chunkSize'], 10)
        self.assertEqual(dimension_data['aerosol_type']['data'][0], 0)
        self.assertEqual(dimension_data['aerosol_type']['data'][-1], 9)

        dimension_data = cci_cdc.get_variable_data(
            OC_DAY_SIN_ID, {'lat': 23761676, 'lon': 23761676}
        )
        self.assertIsNotNone(dimension_data)
        self.assertEqual(dimension_data['lat']['size'], 23761676)
        self.assertEqual(dimension_data['lat']['chunkSize'], 72900)
        self.assertEqual(len(dimension_data['lat']['data']), 0)
        self.assertEqual(dimension_data['lon']['size'], 23761676)
        self.assertEqual(dimension_data['lon']['chunkSize'], 72900)
        self.assertEqual(len(dimension_data['lon']['data']), 0)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_ecv(self):
        cci_cdc = CciOdp()
        aerosol_sources = cci_cdc.search(
            start_date='1990-05-01',
            end_date='2021-08-01',
            bbox=(-20, 30, 20, 50),
            cci_attrs=dict(
                ecv='AEROSOL'
            )
        )
        self.assertTrue(len(aerosol_sources) > 13)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_frequency(self):
        cci_cdc = CciOdp()
        five_day_sources = cci_cdc.search(
            start_date='1990-05-01',
            end_date='2021-08-01',
            bbox=(-20, 30, 20, 50),
            cci_attrs=dict(
                frequency='5 days'
            )
        )
        self.assertTrue(len(five_day_sources) > 10)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_processing_level(self):
        cci_cdc = CciOdp()
        l2p_sources = cci_cdc.search(
            start_date='1990-05-01',
            end_date='2021-08-01',
            bbox=(-20, 30, 20, 50),
            cci_attrs=dict(
                processing_level='L2P'
            )
        )
        self.assertTrue(len(l2p_sources) > 10)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_product_string(self):
        cci_cdc = CciOdp()
        avhrr19g_sources = cci_cdc.search(
            start_date='1990-05-01',
            end_date='2021-08-01',
            bbox=(-20, 30, 20, 50),
            cci_attrs=dict(
                product_string='AVHRR19_G'
            )
        )
        self.assertTrue(len(avhrr19g_sources) > 3)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_product_version(self):
        cci_cdc = CciOdp(data_type="geodataframe")
        v238_sources = cci_cdc.search(
            start_date='1990-05-01',
            end_date='2021-08-01',
            bbox=(-20, 30, 20, 50),
            cci_attrs=dict(
                product_version='v2.3.8'
            )
        )
        self.assertTrue(len(v238_sources) > 2)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_data_type(self):
        cci_cdc = CciOdp()
        siconc_sources = cci_cdc.search(
            start_date='2007-05-01',
            end_date='2009-08-01',
            bbox=(-20, 30, 20, 50),
            cci_attrs=dict(
                data_type='SICONC'
            )
        )
        self.assertTrue(len(siconc_sources) > 3)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_sensor(self):
        cci_cdc = CciOdp(data_type="geodataframe")
        sciamachy_sources = cci_cdc.search(
            start_date='1990-05-01',
            end_date='2021-08-01',
            bbox=(-20, 30, 20, 50),
            cci_attrs=dict(
                sensor='SCIAMACHY'
            )
        )
        self.assertTrue(len(sciamachy_sources) > 2)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_time_ranges_from_data(self):
        cci_cdc = CciOdp()
        first_time_ranges = cci_cdc.get_time_ranges_from_data(
            dataset_name='esacci.OC.5-days.L3S.RRS.multi-sensor.multi-platform.'
            'MERGED.6-0.geographic',
            start_time='1997-09-03T00:00:00',
            end_time='1997-09-10T00:00:00'
        )
        self.assertEqual([(pd.Timestamp('1997-09-03 00:00:00'),
                          pd.Timestamp('1997-09-07 23:59:00')),
                          (pd.Timestamp('1997-09-08 00:00:00'),
                           pd.Timestamp('1997-09-12 23:59:00'))],
                         first_time_ranges)
        self.assertIsNotNone(first_time_ranges)
        second_time_ranges = cci_cdc.get_time_ranges_from_data(
            dataset_name='esacci.OC.5-days.L3S.RRS.multi-sensor.multi-platform.'
            'MERGED.6-0.geographic',
            start_time='1997-09-10T00:00:00',
            end_time='1997-09-30T00:00:00'
        )
        self.assertIsNotNone(second_time_ranges)
        self.assertEqual([(pd.Timestamp('1997-09-08 00:00:00'),
                          pd.Timestamp('1997-09-12 23:59:00')),
                          (pd.Timestamp('1997-09-13 00:00:00'),
                           pd.Timestamp('1997-09-17 23:59:00')),
                          (pd.Timestamp('1997-09-18 00:00:00'),
                           pd.Timestamp('1997-09-22 23:59:00')),
                          (pd.Timestamp('1997-09-23 00:00:00'),
                           pd.Timestamp('1997-09-27 23:59:00')),
                          (pd.Timestamp('1997-09-28 00:00:00'),
                           pd.Timestamp('1997-10-02 23:59:00'))],
                         second_time_ranges)



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any
from typing import Dict
from typing import Hashable
from typing import Mapping
from typing import Optional
from typing import Tuple
import dask.array as da
import numpy as np
import xarray
import warnings

from xcube.core.store import DataDescriptor
from xcube.core.store import DataType
from xcube.core.store import DataTypeLike
from xcube.core.store import VariableDescriptor
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonObjectSchema


class VectorDataCube(xarray.Dataset):
    """A wrapper class around an xarray Dataset to keep it separate."""
    __slots__ = ()

VECTOR_DATA_CUBE_TYPE = DataType(
    VectorDataCube,
    ['vectordatacube', 'xarray.Dataset']
)

DataType.register_data_type(VECTOR_DATA_CUBE_TYPE)


def _attrs_to_json(attrs: Mapping[Hashable, Any]) \
        -> Optional[Dict[str, Any]]:
    new_attrs: Dict[str, Any] = {}
    for k, v in attrs.items():
        if isinstance(v, np.ndarray):
            v = v.tolist()
        elif isinstance(v, da.Array):
            v = np.array(v).tolist()
        if isinstance(v, float) and np.isnan(v):
            v = None
        new_attrs[str(k)] = v
    return new_attrs


class VectorDataCubeDescriptor(DataDescriptor):
    """
    A descriptor for an N-dimensional vector data cube whose spatial properties are
    designated by a geometry cooordinate. The cube is represented by an xarray.Dataset.
    Comprises a description of the data variables contained in the dataset.

    Regarding *time_range* and *time_period* parameters, please refer to
    https://github.com/dcs4cop/xcube/blob/master/docs/source/storeconv.md#date-time-and-duration-specifications

    :param data_id: An identifier for the data
    :param data_type: The data type of the data described
    :param crs: A coordinate reference system identifier,
        as an EPSG, PROJ or WKT string
    :param bbox: A bounding box of the data
    :param time_range: Start and end time delimiting this data's
        temporal extent
    :param time_period: The data's periodicity
        if it is evenly temporally resolved
    :param dims: A mapping of the dataset's dimensions to their sizes
    :param coords: mapping of the dataset's data coordinate names
        to instances of :class:VariableDescriptor
    :param data_vars: A mapping of the dataset's variable names
        to instances of :class:VariableDescriptor
    :param attrs: A mapping containing arbitrary attributes of the dataset
    :param open_params_schema: A JSON schema describing the parameters
        that may be used to open this data
    """

    def __init__(self,
                 data_id: str,
                 *,
                 data_type: DataTypeLike = VECTOR_DATA_CUBE_TYPE,
                 crs: str = None,
                 bbox: Tuple[float, float, float, float] = None,
                 time_range: Tuple[Optional[str], Optional[str]] = None,
                 time_period: str = None,
                 spatial_res: float = None,
                 dims: Mapping[str, int] = None,
                 coords: Mapping[str, 'VariableDescriptor'] = None,
                 data_vars: Mapping[str, 'VariableDescriptor'] = None,
                 attrs: Mapping[Hashable, any] = None,
                 open_params_schema: JsonObjectSchema = None,
                 **additional_properties):
        super().__init__(data_id=data_id,
                         data_type=data_type,
                         crs=crs,
                         bbox=bbox,
                         time_range=time_range,
                         time_period=time_period,
                         open_params_schema=open_params_schema)
        assert_true(VECTOR_DATA_CUBE_TYPE.is_super_type_of(data_type),
                    f'illegal data_type,'
                    f' must be compatible with {VECTOR_DATA_CUBE_TYPE!r}')
        if additional_properties:
            warnings.warn(f'Additional properties received;'
                          f' will be ignored: {additional_properties}')
        self.dims = dict(dims) if dims else None
        self.spatial_res = spatial_res
        self.coords = coords if coords else None
        self.data_vars = data_vars if data_vars else None
        self.attrs = _attrs_to_json(attrs) if attrs else None

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            dims=JsonObjectSchema(
                additional_properties=JsonIntegerSchema(minimum=0)
            ),
            coords=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            data_vars=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            attrs=JsonObjectSchema(
                additional_properties=True
            ),
        )
        schema.required = ['data_id', 'data_type']
        schema.additional_properties = False
        schema.factory = cls
        return schema



import unittest

from xcube_cci.cciodp import CciOdp
from xcube_cci.dataframeaccess import DataFrameAccessor

GHG_DS_ID = "esacci.GHG.satellite-orbit-frequency.L2.CH4.SCIAMACHY.Envisat.IMAP.v7-2.r1"


class DataFrameAccessTest(unittest.TestCase):

    def setUp(self) -> None:
        ccicdc = CciOdp(data_type='geodataframe')
        self._dfa = DataFrameAccessor(ccicdc, GHG_DS_ID, {})

    def test_get_geodataframe(self):
        gdf = self._dfa.get_geodataframe()
        self.assertIsNotNone(gdf)

    def test_get_geodataframe_for_dataset(self):
        gdf = self._dfa._get_features_from_cci_cdc(0)
        self.assertIsNotNone(gdf)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from .constants import COMMON_COORD_VAR_NAMES


def normalize_dims_description(dims: dict) -> dict:
    new_dims = dims.copy()
    if 'latitude' in new_dims:
        new_dims['lat'] = new_dims.pop('latitude')
    if 'longitude' in new_dims:
        new_dims['lon'] = new_dims.pop('longitude')
    if 'latitude_centers' in new_dims:
        new_dims['lat'] = new_dims.pop('latitude_centers')
        new_dims['lon'] = new_dims['lat'] * 2
    return new_dims


def normalize_variable_dims_description(var_dims: List[str]) -> Optional[List[str]]:
    if ('lat' in var_dims and 'lon' in var_dims) or \
            ('latitude' in var_dims and 'longitude' in var_dims) or \
            ('latitude_centers' in var_dims) or \
            ('x' in var_dims and 'y' in var_dims) or \
            ('xc' in var_dims and 'yc' in var_dims):
        if 'y' in var_dims:
            y_dim_name = 'y'
        elif 'yc' in var_dims:
            y_dim_name = 'yc'
        else:
            y_dim_name = 'lat'
        if 'x' in var_dims:
            x_dim_name = 'x'
        elif 'xc' in var_dims:
            x_dim_name = 'xc'
        else:
            x_dim_name = 'lon'
        if var_dims != ['time', y_dim_name, x_dim_name]:
            other_dims = []
            for dim in var_dims:
                if dim not in COMMON_COORD_VAR_NAMES:
                    other_dims.append(dim)
            new_dims = ['time', y_dim_name, x_dim_name]
            for i in range(len(other_dims)):
                new_dims.insert(i + 1, other_dims[i])
            var_dims = new_dims
    return var_dims


def normalize_var_infos(var_infos: Dict[str, Dict[str, Any]]) \
        -> Dict[str, Dict[str, Any]]:
    if 'latitude_centers' in var_infos:
        var_info = var_infos.pop('latitude_centers')
        var_infos['lat'] = dict(data_type=var_info['data_type'],
                                dimensions='lat',
                                units=var_info.get('units'),
                                long_name='latitude',
                                size=var_info.get('size'),
                                shape=var_info.get('shape'),
                                chunk_sizes=var_info.get('chunk_sizes'))
        var_infos['lon'] = dict(data_type=var_info['data_type'],
                                dimensions='lon',
                                units=var_info.get('units'),
                                long_name='longitude',
                                size=var_info.get('size') * 2,
                                shape=[item * 2 for item in
                                       var_info.get('shape', [])],
                                chunk_sizes=[item * 2 for item in
                                       var_info.get('shape', [])])
    return var_infos


def normalize_coord_names(coord_names: List[str])-> List[str]:
    if 'latitude_centers' in coord_names:
        coord_names.remove('latitude_centers')
        coord_names.append('lat')
        coord_names.append('lon')
    return coord_names



# The MIT License (MIT)
# Copyright (c) 2023-2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os.path

from xcube.core.store import get_data_store_class

ReferenceDataStore = get_data_store_class('reference')


class CciKerchunkDataStore(ReferenceDataStore):
    _kc_refs = None

    def __init__(self):
        if self._kc_refs is None:
            kc_refs_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'data/kc_refs.json'
            )
            with open(kc_refs_path, 'r') as fp:
                self._kc_refs = json.load(fp)
        super().__init__(self._kc_refs, target_options=dict(compression=None))



import unittest

from xcube.util.extension import ExtensionRegistry
from xcube_cci.dataaccess import CciOdpDataStore

from xcube_cci.plugin import init_plugin


class PluginTest(unittest.TestCase):

    def test_extensions_exist(self):
        ext_reg = ExtensionRegistry()

        init_plugin(ext_reg)

        self.assertTrue(ext_reg.has_extension('xcube.core.store', 'esa-cdc'))
        self.assertTrue(ext_reg.has_extension('xcube.core.store',
                                              'esa-climate-data-centre')
                        )

    def test_get_extension_short(self):
        ext_reg = ExtensionRegistry()

        init_plugin(ext_reg)

        store = ext_reg.get_component('xcube.core.store', 'esa-cdc')

        self.assertIsNotNone(store)
        self.assertIsInstance(store(), CciOdpDataStore)

    def test_get_extension_long(self):
        ext_reg = ExtensionRegistry()

        init_plugin(ext_reg)

        store = ext_reg.get_component('xcube.core.store',
                                      'esa-climate-data-centre'
                                      )
        self.assertIsNotNone(store)
        self.assertIsInstance(store(), CciOdpDataStore)



import datetime as dt
import os
import unittest
from unittest import skip
from unittest import skipIf

from xcube.core.gridmapping import GridMapping
from xcube.core.normalize import normalize_dataset
from xcube.core.store import DATASET_TYPE
from xcube.core.store import DataStoreError
from xcube.core.store import GEO_DATA_FRAME_TYPE
from xcube.core.store.descriptor import DatasetDescriptor
from xcube.core.store.descriptor import GeoDataFrameDescriptor
from xcube.core.verify import assert_cube

from xcube_cci.dataaccess import CciOdpDataStore
from xcube_cci.dataaccess import CciOdpDataFrameOpener
from xcube_cci.dataaccess import CciOdpDatasetOpener
from xcube_cci.dataaccess import get_temporal_resolution_from_id
from xcube_cci.dataaccess import CciOdpVectorDataCubeOpener
from xcube_cci.dataaccess import VectorDataCubeDescriptor
from xcube_cci.dataaccess import VECTOR_DATA_CUBE_TYPE

AEROSOL_DAY_ID = 'esacci.AEROSOL.day.L3.AAI.multi-sensor.multi-platform.' \
                 'MSAAI.1-7.r1'
AEROSOL_DAY_ENVISAT_ID = 'esacci.AEROSOL.day.L3C.AER_PRODUCTS.AATSR.Envisat.' \
                         'ATSR2-ENVISAT-ENS_DAILY.v2-6.r1'
AEROSOL_CLIMATOLOGY_ID = 'esacci.AEROSOL.climatology.L3.AAI.multi-sensor.' \
                         'multi-platform.MSAAI.1-7.r1'
BIOMASS_ID = "esacci.BIOMASS.yr.L4.AGB.multi-sensor.multi-platform.MERGED.5-0.1000m"
CLOUD_ID = "esacci.CLOUD.mon.L3C.CLD_PRODUCTS.MODIS.Terra.MODIS_TERRA.2-0.r1"
ICESHEETS_ID = 'esacci.ICESHEETS.yr.Unspecified.IV.PALSAR.ALOS.UNSPECIFIED.' \
               '1-1.greenland_margin_2006_2011'
LAKES_ID = 'esacci.LAKES.day.L3S.LK_PRODUCTS.multi-sensor.multi-platform.' \
           'MERGED.v1-1.r1'
OZONE_MON_ID = 'esacci.OZONE.mon.L3.NP.multi-sensor.multi-platform.MERGED.' \
               'fv0002.r1'
OZONE_MON_GOMOS_ID = 'esacci.OZONE.mon.L3.LP.GOMOS.Envisat.GOMOS_ENVISAT.' \
                     'v0001.r1'
OZONE_MON_SCIAMACHY_ID = 'esacci.OZONE.mon.L3.LP.SCIAMACHY.Envisat.' \
                         'SCIAMACHY_ENVISAT.v0001.r1'
SEAICE_ID = 'esacci.SEAICE.day.L4.SICONC.multi-sensor.multi-platform.' \
            'AMSR_25kmEASE2.2-1.NH'
SST_ID = 'esacci.SST.day.L4.SSTdepth.multi-sensor.multi-platform.OSTIA.1-1.r1'
GHG_DS_ID = "esacci.GHG.satellite-orbit-frequency.L2.CH4.SCIAMACHY.Envisat.IMAP.v7-2.r1"
VDC_ID = "esacci.SEALEVEL.mon.IND.MSLTR.multi-sensor.multi-platform.MERGED.2-2.WAFRICA"
COORDS_2D_ID = "esacci.ICESHEETS.unspecified.L4.SEC.multi-sensor.multi-platform." \
               "UNSPECIFIED.0-1.greenland_sec_saral_altika"


class DataAccessTest(unittest.TestCase):

    def test_get_temporal_resolution_from_id(self):
        self.assertEqual('1D', get_temporal_resolution_from_id(
            'esacci.OZONE.day.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('5D', get_temporal_resolution_from_id(
            'esacci.OZONE.5-days.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('1M', get_temporal_resolution_from_id(
            'esacci.OZONE.mon.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('1M', get_temporal_resolution_from_id(
            'esacci.OZONE.month.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('3M', get_temporal_resolution_from_id(
            'esacci.OZONE.3-months.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('1Y', get_temporal_resolution_from_id(
            'esacci.OZONE.yr.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('1Y', get_temporal_resolution_from_id(
            'esacci.OZONE.year.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('13Y', get_temporal_resolution_from_id(
            'esacci.OZONE.13-yrs.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertEqual('1M', get_temporal_resolution_from_id(
            'esacci.OZONE.climatology.L3.NP.sensor.platform.MERGED.fv0002.r1'
        ))
        self.assertIsNone(get_temporal_resolution_from_id(
            'esacci.OZONE.satellite-orbit-frequency.L3.NP.sensor.platform.'
            'MERGED.fv0002.r1'
        ))


class CciOdpDatasetOpenerTest(unittest.TestCase):

    def setUp(self) -> None:
        self.opener = CciOdpDatasetOpener(normalize_data=False)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_dataset_names(self):
        self.assertTrue(len(self.opener.dataset_names) > 275)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_has_data(self):
        self.assertTrue(self.opener.has_data(
            'esacci.FIRE.mon.L4.BA.MODIS.Terra.MODIS_TERRA.v5-1.grid')
        )
        self.assertFalse(self.opener.has_data(
            'esacci.WIND.mon.L4.BA.MODIS.Terra.MODIS_TERRA.v5-1.grid')
        )
        self.assertTrue(self.opener.has_data(
            'esacci.AEROSOL.day.L3.AAI.multi-sensor.multi-platform.MSAAI.'
            '1-7.r1')
        )
        self.assertTrue(self.opener.has_data(
            'esacci.OZONE.mon.L3.NP.multi-sensor.multi-platform.MERGED.'
            'fv0002.r1')
        )

        self.assertTrue(self.opener.has_data(
            'esacci.FIRE.mon.L4.BA.MODIS.Terra.MODIS_TERRA.v5-1.grid')
        )
        self.assertFalse(self.opener.has_data(
            'esacci.WIND.mon.L4.BA.MODIS.Terra.MODIS_TERRA.v5-1.grid')
        )
        self.assertTrue(self.opener.has_data(
            'esacci.AEROSOL.day.L3.AAI.multi-sensor.multi-platform.MSAAI.'
            '1-7.r1')
        )
        self.assertTrue(self.opener.has_data(
            'esacci.OZONE.mon.L3.NP.multi-sensor.multi-platform.MERGED.'
            'fv0002.r1')
        )

        self.assertTrue(self.opener.has_data(
            'esacci.FIRE.mon.L4.BA.MODIS.Terra.MODIS_TERRA.v5-1.grid')
        )
        self.assertFalse(self.opener.has_data(
            'esacci.WIND.mon.L4.BA.MODIS.Terra.MODIS_TERRA.v5-1.grid')
        )
        self.assertTrue(self.opener.has_data(
            'esacci.AEROSOL.day.L3.AAI.multi-sensor.multi-platform.MSAAI.'
            '1-7.r1')
        )
        self.assertTrue(self.opener.has_data(
            'esacci.OZONE.mon.L3.NP.multi-sensor.multi-platform.MERGED.'
            'fv0002.r1')
        )

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_dataset_crs_variable(self):
        descriptor = self.opener.describe_data(SEAICE_ID)
        self.assertIsNotNone(descriptor)
        self.assertTrue('Lambert_Azimuthal_Grid' in
                        descriptor.data_vars.keys())
        self.assertTrue('xc' in descriptor.coords.keys())
        self.assertTrue('yc' in descriptor.coords.keys())
        self.assertEqual('Lambert Azimuthal Equal Area', descriptor.crs)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_climatology_dataset(self):
        descriptor = self.opener.describe_data(AEROSOL_CLIMATOLOGY_ID)
        self.assertIsNotNone(descriptor)
        self.assertTrue('absorbing_aerosol_index' in
                        descriptor.data_vars.keys())
        self.assertEqual(
            {'longitude', 'latitude', 'month'}, set(descriptor.coords.keys())
        )
        self.assertIsNone(descriptor.time_range)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_data(self):
        descriptor = self.opener.describe_data(OZONE_MON_ID)
        self.assertIsNotNone(descriptor)
        self.assertIsInstance(descriptor, DatasetDescriptor)
        self.assertEqual(OZONE_MON_ID, descriptor.data_id)
        self.assertEqual('dataset', str(descriptor.data_type))
        self.assertEqual(['lon', 'lat', 'layers', 'air_pressure', 'time',
                          'bnds'],
                         list(descriptor.dims.keys()))
        self.assertEqual(360, descriptor.dims['lon'])
        self.assertEqual(180, descriptor.dims['lat'])
        self.assertEqual(16, descriptor.dims['layers'])
        self.assertEqual(17, descriptor.dims['air_pressure'])
        self.assertEqual(36, descriptor.dims['time'])
        self.assertEqual(2, descriptor.dims['bnds'])
        self.assertEqual(9, len(descriptor.data_vars))
        self.assertTrue('surface_pressure' in descriptor.data_vars)
        self.assertEqual(3, descriptor.data_vars['surface_pressure'].ndim)
        self.assertEqual(('time', 'lat', 'lon'),
                         descriptor.data_vars['surface_pressure'].dims)
        self.assertEqual('float32',
                         descriptor.data_vars['surface_pressure'].dtype)
        self.assertEqual('WGS84', descriptor.crs)
        self.assertEqual(1.0, descriptor.spatial_res)
        self.assertEqual(('1997-01-01', '2008-12-31'),
                         descriptor.time_range)
        self.assertEqual('1M', descriptor.time_period)

        descriptor = self.opener.describe_data(AEROSOL_DAY_ID)
        self.assertIsNotNone(descriptor)
        self.assertIsInstance(descriptor, DatasetDescriptor)
        self.assertEqual(AEROSOL_DAY_ID, descriptor.data_id)
        self.assertEqual('dataset', str(descriptor.data_type))
        self.assertEqual(['longitude', 'latitude', 'time', 'bnds'],
                         list(descriptor.dims.keys()))
        self.assertEqual(360, descriptor.dims['longitude'])
        self.assertEqual(180, descriptor.dims['latitude'])
        self.assertEqual(12644, descriptor.dims['time'])
        self.assertEqual(2, descriptor.dims['bnds'])
        self.assertEqual(3, len(descriptor.data_vars))
        self.assertTrue('absorbing_aerosol_index' in descriptor.data_vars)
        self.assertEqual(3,
                         descriptor.data_vars['absorbing_aerosol_index'].ndim)
        self.assertEqual(('time', 'latitude', 'longitude'),
                         descriptor.data_vars['absorbing_aerosol_index'].dims)
        self.assertEqual('float32',
                         descriptor.data_vars['absorbing_aerosol_index'].dtype)
        self.assertEqual('WGS84', descriptor.crs)
        self.assertEqual(1.0, descriptor.spatial_res)
        self.assertEqual(('1978-11-01', '2015-12-31'),
                         descriptor.time_range)
        self.assertEqual('1D', descriptor.time_period)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_2d_grid_coords_data(self):
        descriptor = self.opener.describe_data(COORDS_2D_ID)
        self.assertIsNotNone(descriptor)
        self.assertEqual('dataset', str(descriptor.data_type))
        self.assertEqual(['Time', 'x', 'y', 'bnds'], list(descriptor.dims.keys()))
        self.assertEqual(634, descriptor.dims['x'])
        self.assertEqual(635, descriptor.dims['y'])
        self.assertEqual(1, descriptor.dims['Time'])
        self.assertEqual(2, descriptor.dims['bnds'])
        self.assertEqual(7, len(descriptor.data_vars))
        self.assertTrue('CS2SEC' in descriptor.data_vars)
        self.assertEqual(3, descriptor.data_vars['CS2SEC'].ndim)
        self.assertEqual(('y', 'x', 'Time'),
                         descriptor.data_vars['CS2SEC'].dims)
        self.assertEqual('float32',
                         descriptor.data_vars['CS2SEC'].dtype)
        self.assertEqual('Polar Stereographic (variant B)', descriptor.crs)
        self.assertIsNone(descriptor.spatial_res)
        self.assertEqual(('2013-03-31', '2017-03-31'), descriptor.time_range)
        self.assertIsNone(descriptor.time_period)

    def test_get_open_data_params_schema_no_data(self):
        schema = self.opener.get_open_data_params_schema().to_dict()
        self.assertIsNotNone(schema)
        self.assertTrue('variable_names' in schema['properties'])
        self.assertTrue('time_range' in schema['properties'])
        self.assertFalse(schema['additionalProperties'])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_open_data_params_schema(self):
        schema = self.opener.get_open_data_params_schema(OZONE_MON_ID).to_dict()
        self.assertIsNotNone(schema)
        self.assertTrue('variable_names' in schema['properties'])
        self.assertTrue('time_range' in schema['properties'])
        self.assertTrue('bbox' in schema['properties'])
        self.assertFalse(schema['additionalProperties'])

        schema = self.opener.get_open_data_params_schema(AEROSOL_DAY_ID).\
            to_dict()
        self.assertIsNotNone(schema)
        self.assertTrue('variable_names' in schema['properties'])
        self.assertTrue('time_range' in schema['properties'])
        self.assertTrue('bbox' in schema['properties'])
        self.assertFalse(schema['additionalProperties'])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_data_sst(self):
        dataset = self.opener.open_data(
            SST_ID,
            variable_names=['analysed_sst'],
            bbox=[-30, 20, -29, 21],
            time_range=['2008-07-01','2008-07-01'])

        self.assertIsNotNone(dataset)
        self.assertEqual({'analysed_sst'}, set(dataset.data_vars))
        self.assertEqual({'lat', 'lat_bnds', 'lon', 'lon_bnds',
                          'time', 'time_bnds'},
                         set(dataset.coords))
        self.assertEqual({'time', 'lat', 'lon'}, set(dataset.analysed_sst.dims))
        self.assertEqual({1, 20, 20}, set(dataset.analysed_sst.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_data_aerosol(self):
        dataset = self.opener.open_data(
            AEROSOL_DAY_ID,
            variable_names=['absorbing_aerosol_index'],
            bbox=[40, 40, 50, 50],
            time_range=['2014-01-01','2014-04-30'])
        self.assertIsNotNone(dataset)
        self.assertEqual({'absorbing_aerosol_index'}, set(dataset.data_vars))
        self.assertEqual({'longitude', 'latitude', 'time', 'time_bnds'},
                         set(dataset.coords))
        self.assertEqual({'time', 'latitude', 'longitude'},
                         set(dataset.absorbing_aerosol_index.dims))
        self.assertEqual({1, 10, 10},
                         set(dataset.absorbing_aerosol_index.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())

        # open same dataset again to ensure chunking is different
        full_dataset = self.opener.open_data(
            AEROSOL_DAY_ID,
            variable_names=['absorbing_aerosol_index'],
            time_range=['2014-01-01','2014-04-30'])
        self.assertIsNotNone(full_dataset)
        self.assertEqual({'absorbing_aerosol_index'},
                         set(full_dataset.data_vars))
        self.assertEqual({'longitude', 'latitude', 'time', 'time_bnds'},
                         set(full_dataset.coords))
        self.assertEqual({'time', 'latitude', 'longitude'},
                         set(full_dataset.absorbing_aerosol_index.dims))
        self.assertEqual({1, 180, 360},
                         set(full_dataset.absorbing_aerosol_index.chunk_sizes))
        self.assertIsNotNone(full_dataset.zarr_store.get())

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_climatology_data(self):
        dataset = self.opener.open_data(AEROSOL_CLIMATOLOGY_ID)
        self.assertIsNotNone(dataset)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_biomass_data(self):
        dataset = self.opener.open_data(BIOMASS_ID)
        self.assertIsNotNone(dataset)
        self.assertIn("time", dataset.coords)
        self.assertEqual(8, len(dataset["time"].values))

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_data_with_crs(self):
        dataset = self.opener.open_data(ICESHEETS_ID)
        self.assertIsNotNone(dataset)
        self.assertTrue('crs' in dataset.data_vars.keys())
        self.assertTrue('grid_mapping_name' in dataset['crs'].attrs)

        grid_mapping = GridMapping.from_dataset(dataset)
        self.assertEqual(('x', 'y'), grid_mapping.xy_var_names)
        self.assertEqual((3200, 5400), grid_mapping.size)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    @skip('Disabled while time series are not supported')
    def test_open_data_with_time_series_and_latitude_centers(self):
        dataset = self.opener.open_data(
            OZONE_MON_SCIAMACHY_ID,
            variable_names=['approximate_altitude', 'ozone_mixing_ratio',
                            'sample_standard_deviation'],
            time_range=['2009-05-02', '2009-08-31'],
            bbox=[-10.0, 40.0, 10.0, 60.0]
        )
        self.assertIsNotNone(dataset)
        self.assertEqual({'approximate_altitude', 'ozone_mixing_ratio',
                          'sample_standard_deviation'},
                         set(dataset.data_vars))
        self.assertEqual({'time', 'air_pressure', 'latitude_centers'},
                         dataset.ozone_mixing_ratio.dims)
        self.assertEqual({1, 32, 18}, dataset.ozone_mixing_ratio.chunk_sizes)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_2d_grid_coords_data(self):
        dataset = self.opener.open_data(
            COORDS_2D_ID,
        )
        self.assertIsNotNone(dataset)
        self.assertEqual(
            {'AltiKaSEC', 'AltiKaSEC_error', 'CS2SEC', 'CS2SEC_error',
             'grid_projection', 'lat', 'lon'},
            set(dataset.data_vars)
        )
        self.assertEqual(('y', 'x', 'Time'), dataset.AltiKaSEC.dims)
        self.assertEqual([635, 634, 1], dataset.AltiKaSEC.chunk_sizes)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search(self):
        search_result = list(self.opener.search_data(cci_attrs=dict(
            ecv='CLOUD',
            product_string='MODIS_TERRA'))
        )
        self.assertIsNotNone(search_result)
        self.assertEqual(1, len(search_result))
        self.assertIsInstance(search_result[0], DatasetDescriptor)
        self.assertEqual(22, len(search_result[0].dims))
        self.assertEqual(147, len(search_result[0].data_vars))
        self.assertEqual(CLOUD_ID, search_result[0].data_id)
        self.assertEqual('1M', search_result[0].time_period)
        self.assertEqual(0.5, search_result[0].spatial_res)
        self.assertEqual(DATASET_TYPE, search_result[0].data_type)
        self.assertEqual(('2000-02-01', '2014-12-31'), search_result[0].time_range)
        self.assertEqual('dataset', search_result[0].data_type.alias)


class CciOdpDatasetOpenerTimeSeriesTest(unittest.TestCase):

    def setUp(self) -> None:
        self.opener = CciOdpDatasetOpener(normalize_data=False)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_monthly_ozone(self):
        descriptor = self.opener.describe_data(OZONE_MON_GOMOS_ID)
        self.assertIsNotNone(descriptor)
        self.assertEqual('1M', descriptor.time_period)
        self.assertEqual(('2002-01-01', '2011-12-31'), descriptor.time_range)
        self.assertEqual({'time', 'air_pressure', 'latitude_centers', 'bnds'},
                         set(descriptor.dims))
        self.assertEqual(120, descriptor.dims['time'])
        self.assertTrue('time' in descriptor.coords)
        self.assertEqual(120, descriptor.coords['time'].attrs.get('size'))

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_monthly_ozone_full(self):
        dataset = self.opener.open_data(
            OZONE_MON_GOMOS_ID,
            variable_names=['approximate_altitude',
                            'ozone_mixing_ratio',
                            'sample_standard_deviation']
        )
        self.assertIsNotNone(dataset)
        self.assertEqual({'approximate_altitude', 'ozone_mixing_ratio',
                          'sample_standard_deviation'},
                         set(dataset.data_vars))
        self.assertEqual({'time', 'air_pressure', 'latitude_centers'},
                         set(dataset.ozone_mixing_ratio.dims))
        self.assertEqual({120, 51, 18},
                         set(dataset.ozone_mixing_ratio.shape))
        self.assertEqual({12, 51, 18},
                         set(dataset.ozone_mixing_ratio.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_monthly_ozone_time_restricted(self):
        dataset = self.opener.open_data(
            OZONE_MON_GOMOS_ID,
            variable_names=['approximate_altitude',
                            'ozone_mixing_ratio',
                            'sample_standard_deviation'],
            time_range=['2004-05-01', '2004-08-31']
        )
        self.assertIsNotNone(dataset)
        self.assertEqual({'approximate_altitude', 'ozone_mixing_ratio',
                          'sample_standard_deviation'},
                         set(dataset.data_vars))
        self.assertEqual({'time', 'air_pressure', 'latitude_centers'},
                         set(dataset.ozone_mixing_ratio.dims))
        self.assertEqual({12, 51, 18},
                         set(dataset.ozone_mixing_ratio.shape))
        self.assertEqual({12, 51, 18},
                         set(dataset.ozone_mixing_ratio.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())


class CciOdpDatasetOpenerTimeSeriesNormalizeTest(unittest.TestCase):

    def setUp(self) -> None:
        self.opener = CciOdpDatasetOpener(normalize_data=True)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_monthly_ozone(self):
        descriptor = self.opener.describe_data(OZONE_MON_GOMOS_ID)
        self.assertIsNotNone(descriptor)
        self.assertEqual('1M', descriptor.time_period)
        self.assertEqual(('2002-01-01', '2011-12-31'), descriptor.time_range)
        self.assertEqual({'time', 'air_pressure', 'lat', 'lon', 'bnds'},
                         set(descriptor.dims))
        self.assertEqual(120, descriptor.dims['time'])
        self.assertTrue('time' in descriptor.coords)
        self.assertEqual(120, descriptor.coords['time'].attrs.get('size'))

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_monthly_ozone_full(self):
        dataset = self.opener.open_data(
            OZONE_MON_GOMOS_ID,
            variable_names=['approximate_altitude',
                            'ozone_mixing_ratio',
                            'sample_standard_deviation']
        )
        self.assertIsNotNone(dataset)
        self.assertEqual({'approximate_altitude', 'ozone_mixing_ratio',
                          'sample_standard_deviation'},
                         set(dataset.data_vars))
        self.assertEqual({'time', 'air_pressure', 'lat', 'lon'},
                         set(dataset.ozone_mixing_ratio.dims))
        self.assertEqual({120, 51, 18, 36},
                         set(dataset.ozone_mixing_ratio.shape))
        self.assertEqual({12, 51, 18, 36},
                         set(dataset.ozone_mixing_ratio.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_monthly_ozone_time_restricted(self):
        dataset = self.opener.open_data(
            OZONE_MON_GOMOS_ID,
            variable_names=['approximate_altitude',
                            'ozone_mixing_ratio',
                            'sample_standard_deviation'],
            time_range=['2004-05-01', '2004-08-31']
        )
        self.assertIsNotNone(dataset)
        self.assertEqual({'approximate_altitude', 'ozone_mixing_ratio',
                          'sample_standard_deviation'},
                         set(dataset.data_vars))
        self.assertEqual({'time', 'air_pressure', 'lat', 'lon'},
                         set(dataset.ozone_mixing_ratio.dims))
        self.assertEqual({12, 51, 18, 36},
                         set(dataset.ozone_mixing_ratio.shape))
        self.assertEqual({12, 51, 18, 36},
                         set(dataset.ozone_mixing_ratio.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())


class CciOdpDatasetOpenerNormalizeTest(unittest.TestCase):

    def setUp(self) -> None:
        self.opener = CciOdpDatasetOpener(normalize_data=True)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_dataset_crs_variable(self):
        descriptor = self.opener.describe_data(ICESHEETS_ID)
        self.assertIsNotNone(descriptor)
        self.assertTrue('crs' in descriptor.data_vars.keys())

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_dataset(self):
        descriptor = self.opener.describe_data(AEROSOL_DAY_ID)
        self.assertIsNotNone(descriptor)
        self.assertIsInstance(descriptor, DatasetDescriptor)
        self.assertEqual(AEROSOL_DAY_ID, descriptor.data_id)
        self.assertEqual('dataset', str(descriptor.data_type))
        self.assertEqual({'lat', 'lon', 'time', 'bnds'},
                         set(descriptor.dims.keys()))
        self.assertEqual(360, descriptor.dims['lon'])
        self.assertEqual(180, descriptor.dims['lat'])
        self.assertEqual(12644, descriptor.dims['time'])
        self.assertEqual(2, descriptor.dims['bnds'])
        self.assertEqual(3, len(descriptor.data_vars))
        self.assertTrue('absorbing_aerosol_index' in descriptor.data_vars)
        self.assertEqual(3,
                         descriptor.data_vars['absorbing_aerosol_index'].ndim)
        self.assertEqual(('time', 'lat', 'lon'),
                         descriptor.data_vars['absorbing_aerosol_index'].dims)
        self.assertEqual('float32',
                         descriptor.data_vars['absorbing_aerosol_index'].dtype)
        self.assertEqual('WGS84', descriptor.crs)
        self.assertEqual(1.0, descriptor.spatial_res)
        self.assertEqual(('1978-11-01', '2015-12-31'), descriptor.time_range)
        self.assertEqual('1D', descriptor.time_period)

    def test_get_open_data_params_schema_no_data(self):
        schema = self.opener.get_open_data_params_schema().to_dict()
        self.assertIsNotNone(schema)
        self.assertTrue('variable_names' in schema['properties'])
        self.assertTrue('time_range' in schema['properties'])
        self.assertFalse(schema['additionalProperties'])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_open_data_params_schema(self):
        schema = self.opener.get_open_data_params_schema(AEROSOL_DAY_ID).\
            to_dict()
        self.assertIsNotNone(schema)
        self.assertTrue('variable_names' in schema['properties'])
        self.assertTrue('time_range' in schema['properties'])
        self.assertTrue('bbox' in schema['properties'])
        self.assertFalse(schema['additionalProperties'])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_data(self):
        with self.assertRaises(DataStoreError) as dse:
            self.opener.open_data(AEROSOL_DAY_ENVISAT_ID,
                                  variable_names=['AOD550', 'NMEAS'],
                                  time_range=['2009-07-02', '2009-07-05'],
                                  bbox=[-10.0, 40.0, 10.0, 60.0])
        self.assertEqual(
            f'Cannot describe metadata of data resource '
            f'"{AEROSOL_DAY_ENVISAT_ID}", as it cannot be accessed by '
            f'data accessor "dataset:zarr:cciodp".', f'{dse.exception}')

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_data_aerosol(self):
        dataset = self.opener.open_data(
            AEROSOL_DAY_ID,
            variable_names=['absorbing_aerosol_index'],
            bbox=[40, 40, 50, 50],
            time_range=['2014-01-01','2014-04-30'])
        self.assertIsNotNone(dataset)
        self.assertEqual({'absorbing_aerosol_index'}, set(dataset.data_vars))
        self.assertEqual({'lon', 'lat', 'time', 'time_bnds'},
                         set(dataset.coords))
        self.assertEqual({'time', 'lat', 'lon'},
                         set(dataset.absorbing_aerosol_index.dims))
        self.assertEqual({1, 10, 10},
                         set(dataset.absorbing_aerosol_index.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())

        # open same dataset again to ensure chunking is different
        full_dataset = self.opener.open_data(
            AEROSOL_DAY_ID,
            variable_names=['absorbing_aerosol_index'],
            time_range=['2014-01-01','2014-04-30'])
        self.assertIsNotNone(full_dataset)
        self.assertEqual({'absorbing_aerosol_index'},
                         set(full_dataset.data_vars))
        self.assertEqual({'lon', 'lat', 'time', 'time_bnds'},
                         set(full_dataset.coords))
        self.assertEqual({'time', 'lat', 'lon'},
                         set(full_dataset.absorbing_aerosol_index.dims))
        self.assertEqual({1, 180, 360},
                         set(full_dataset.absorbing_aerosol_index.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    @skip('Disabled while time series are not supported')
    def test_open_data_with_time_series_and_latitude_centers(self):
        dataset = self.opener.open_data(
            OZONE_MON_SCIAMACHY_ID,
            variable_names=['standard_error_of_the_mean', 'ozone_mixing_ratio',
                            'sample_standard_deviation'],
            time_range=['2009-05-02', '2009-08-31'],
            bbox=[-10.0, 40.0, 10.0, 60.0]
        )
        self.assertIsNotNone(dataset)
        self.assertEqual({'standard_error_of_the_mean', 'ozone_mixing_ratio',
                          'sample_standard_deviation'}, set(dataset.data_vars))
        self.assertEqual({'time', 'air_pressure', 'lat', 'lon'},
                         dataset.ozone_mixing_ratio.dims)
        self.assertEqual({1, 32, 18, 36},
                         set(dataset.ozone_mixing_ratio.chunk_sizes))
        self.assertIsNotNone(dataset.zarr_store.get())

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search(self):
        search_result = list(self.opener.search_data(
            cci_attrs=(dict(ecv='LAKES')))
        )
        self.assertIsNotNone(search_result)
        self.assertEqual(4, len(search_result))
        self.assertIsInstance(search_result[1], DatasetDescriptor)
        self.assertEqual(4, len(search_result[1].dims))
        self.assertEqual(52, len(search_result[1].data_vars))
        self.assertEqual(LAKES_ID, search_result[1].data_id)
        self.assertEqual('1D', search_result[1].time_period)
        self.assertEqual(0.008333333, search_result[1].spatial_res)
        self.assertEqual(DATASET_TYPE, search_result[1].data_type)
        self.assertEqual(('1992-09-15', '2019-12-31'),
                         search_result[1].time_range)
        self.assertEqual('dataset', search_result[0].data_type.alias)
        self.assertEqual('dataset', search_result[1].data_type.alias)
        self.assertEqual('dataset', search_result[2].data_type.alias)


def user_agent(ext: str = "") -> str:
    from xcube_cci.version import version
    from platform import machine, python_version, system

    return "xcube-cci-testing/{version} (Python {python};" \
           " {system} {arch}) {ext}".format(
        version=version,
        python=python_version(),
        system=system(),
        arch=machine(),
        ext=ext
    )


class CciOdpDataFrameOpenerTest(unittest.TestCase):

    def setUp(self) -> None:
        self._opener = CciOdpDataFrameOpener()

    def test_dataset_names(self):
        ds_names = self._opener.dataset_names
        self.assertTrue(len(ds_names) > 10)

    def test_get_dataset_types(self):
        data_types = self._opener.get_data_types()
        self.assertEqual(1, len(data_types))
        self.assertEqual(GEO_DATA_FRAME_TYPE, data_types[0])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_has_data(self):
        self.assertTrue(self._opener.has_data(GHG_DS_ID))

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_data(self):
        descriptor = self._opener.describe_data(GHG_DS_ID)
        self.assertIsNotNone(descriptor)
        self.assertIsInstance(descriptor, GeoDataFrameDescriptor)
        self.assertEqual(GHG_DS_ID, descriptor.data_id)
        self.assertEqual("WGS84", descriptor.crs)
        self.assertAlmostEqual(-180, descriptor.bbox[0], 1)
        self.assertAlmostEqual(-90, descriptor.bbox[1], 1)
        self.assertAlmostEqual(180, descriptor.bbox[2], 1)
        self.assertAlmostEqual(90, descriptor.bbox[3], 1)
        self.assertEqual("2003-01-08", descriptor.time_range[0])
        self.assertEqual("2012-04-08", descriptor.time_range[1])
        var_list = {
            'solar_zenith_angle', 'sensor_zenith_angle', 'time', 'pressure_levels',
            'pressure_weight', 'xch4_raw', 'xch4', 'h2o_ecmwf', 'xch4_prior',
            'xco2_prior', 'xco2_retrieved', 'xch4_uncertainty', 'xch4_averaging_kernel',
            'ch4_profile_apriori', 'xch4_quality_flag', 'dry_airmass_layer',
            'surface_elevation', 'surface_temperature', 'chi2_ch4', 'chi2_co2',
            'xco2_macc', 'xco2_CT2015', 'xch4_v71', 'geometry'
        }
        self.assertEqual(
            var_list, set(descriptor.feature_schema.properties.keys())
        )

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_open_data_params_schema(self):
        schema = self._opener.get_open_data_params_schema(GHG_DS_ID).to_dict()
        self.assertIsNotNone(schema)
        self.assertEqual(
            ["variable_names", "time_range", "bbox"],
            list(schema.get("properties").keys())
        )
        var_list = {
            'solar_zenith_angle', 'sensor_zenith_angle', 'pressure_levels',
            'pressure_weight', 'xch4_raw', 'xch4', 'h2o_ecmwf', 'xch4_prior',
            'xco2_prior', 'xco2_retrieved', 'xch4_uncertainty', 'xch4_averaging_kernel',
            'ch4_profile_apriori', 'xch4_quality_flag', 'dry_airmass_layer',
            'surface_elevation', 'surface_temperature', 'chi2_ch4', 'chi2_co2',
            'xco2_macc', 'xco2_CT2015', 'xch4_v71'
        }
        self.assertEqual(
            var_list,
            set(
                schema.get("properties").get("variable_names").get("items").get("enum")
            )
        )


class CciOdpVectorDataCubeOpenerTest(unittest.TestCase):

    def setUp(self) -> None:
        self._opener = CciOdpVectorDataCubeOpener()

    def test_dataset_names(self):
        ds_names = self._opener.dataset_names
        self.assertTrue(len(ds_names) > 10)

    def test_get_data_types(self):
        data_types = self._opener.get_data_types()
        self.assertEqual(1, len(data_types))
        self.assertEqual(VECTOR_DATA_CUBE_TYPE, data_types[0])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_has_data(self):
        self.assertTrue(self._opener.has_data(VDC_ID))

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_describe_data_sealevel(self):
        descriptor = self._opener.describe_data(VDC_ID)
        self.assertIsNotNone(descriptor)
        self.assertIsInstance(descriptor, VectorDataCubeDescriptor)
        self.assertEqual(VDC_ID, descriptor.data_id)
        self.assertEqual('vectordatacube', str(descriptor.data_type))
        self.assertEqual(['nbpoints', 'nbmonth'], list(descriptor.dims.keys()))
        self.assertEqual(1867, descriptor.dims['nbpoints'])
        self.assertEqual(216, descriptor.dims['nbmonth'])
        self.assertEqual(4, len(descriptor.data_vars))
        self.assertTrue('distance_to_coast' in descriptor.data_vars)
        self.assertEqual(1, descriptor.data_vars['distance_to_coast'].ndim)
        self.assertEqual(tuple(['nbpoints']),
                         descriptor.data_vars['distance_to_coast'].dims)
        self.assertEqual('float32', descriptor.data_vars['distance_to_coast'].dtype)
        self.assertEqual('WGS84', descriptor.crs)
        self.assertEqual(('2002-01-01', '2019-12-31'), descriptor.time_range)
        self.assertEqual('1M', descriptor.time_period)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_open_data_params_schema(self):
        schema = self._opener.get_open_data_params_schema(VDC_ID).to_dict()
        self.assertIsNotNone(schema)
        self.assertTrue('variable_names' in schema['properties'])
        self.assertFalse(schema['additionalProperties'])

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_open_data(self):
        data = self._opener.open_data(VDC_ID,
                                      variable_names=["distance_to_coast", "sla"])
        self.assertIsNotNone(data)
        self.assertEqual({"distance_to_coast", "sla"}, set(data.data_vars))
        self.assertEqual({"geometry", "nbmonth", "nbmonth_bnds"}, set(data.coords))
        self.assertEqual({"nbpoints"}, set(data.distance_to_coast.dims))
        self.assertEqual({50}, set(data.distance_to_coast.chunk_sizes))
        self.assertEqual({1867}, set(data.distance_to_coast.shape))
        self.assertEqual({"nbpoints", "nbmonth"}, set(data.sla.dims))
        self.assertEqual({50, 216}, set(data.sla.chunk_sizes))
        self.assertEqual({1867, 216}, set(data.sla.shape))
        self.assertIsNotNone(data.zarr_store.get())


class CciOdpDataStoreTest(unittest.TestCase):

    def setUp(self) -> None:
        odp_params = {'user_agent': user_agent()}
        self.store = CciOdpDataStore(normalize_data=False, **odp_params)

    def test_get_data_store_params_schema(self):
        cci_store_params_schema = \
            CciOdpDataStore.get_data_store_params_schema().to_dict()
        self.assertIsNotNone(cci_store_params_schema)
        self.assertTrue('endpoint_url' in cci_store_params_schema['properties'])
        self.assertTrue(
            'endpoint_description_url' in cci_store_params_schema['properties']
        )
        self.assertTrue('user_agent' in cci_store_params_schema['properties'])

    def test_get_data_types(self):
        self.assertEqual(
            ('dataset', "geodataframe", "vectordatacube"),
            CciOdpDataStore.get_data_types()
        )

    def test_get_data_types_for_data(self):
        data_types_for_data = self.store.get_data_types_for_data(
            AEROSOL_DAY_ID)
        self.assertEqual(('dataset',), data_types_for_data)

        data_types_for_data = self.store.get_data_types_for_data(OZONE_MON_ID)
        self.assertEqual(('dataset',), data_types_for_data)

        with self.assertRaises(DataStoreError) as dse:
            self.store.get_data_types_for_data('nonsense')
        self.assertEqual("Data resource 'nonsense' does not exist in store",
                         f'{dse.exception}')

    def test_get_search_params(self):
        search_schema = self.store.get_search_params_schema().to_dict()
        self.assertIsNotNone(search_schema)
        self.assertTrue('start_date' in search_schema['properties'])
        self.assertTrue('end_date' in search_schema['properties'])
        self.assertTrue('bbox' in search_schema['properties'])
        self.assertTrue('cci_attrs' in search_schema['properties'])
        cci_properties = search_schema['properties']['cci_attrs']['properties']
        self.assertTrue('ecv' in cci_properties)
        self.assertTrue('frequency' in cci_properties)
        self.assertTrue('institute' in cci_properties)
        self.assertTrue('processing_level' in cci_properties)
        self.assertTrue('product_string' in cci_properties)
        self.assertTrue('product_version' in cci_properties)
        self.assertTrue('data_type' in cci_properties)
        self.assertTrue('sensor' in cci_properties)
        self.assertTrue('platform' in cci_properties)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_geodataframe(self):
        geodataframe_search_result = \
            list(self.store.search_data('geodataframe'))
        self.assertIsNotNone(geodataframe_search_result)
        self.assertTrue(len(geodataframe_search_result) > 20)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_search_vectordatacube(self):
        vectordatacube_search_result = list(self.store.search_data('vectordatacube'))
        self.assertIsNotNone(vectordatacube_search_result)
        self.assertTrue(len(vectordatacube_search_result) > 10)


class CciDataNormalizationTest(unittest.TestCase):

    @skip('Execute to test whether all data sets can be normalized')
    def test_normalization(self):
        store = CciOdpDataStore(normalize_data=True)
        all_data = store.search_data()
        datasets_without_variables = []
        datasets_with_unsupported_frequencies = []
        datasets_that_could_not_be_opened = {}
        good_datasets = []
        for data in all_data:
            if 'climatology' in data.data_id:
                print(f'Cannot read data due to unsupported frequency')
                datasets_with_unsupported_frequencies.append(data.data_id)
                continue
            if not data.data_vars or len(data.data_vars) < 1:
                print(f'Dataset {data.data_id} '
                      f'has not enough variables to open. Will skip.')
                datasets_without_variables.append(data.data_id)
                continue
            print(f'Attempting to open {data.data_id} ...')
            variable_names = []
            for i in range(min(3, len(data.data_vars))):
                variable_names.append(list(data.data_vars.keys())[i])
            start_time = dt.datetime.strptime(data.time_range[0],
                                              '%Y-%m-%d').timestamp()
            end_time = dt.datetime.strptime(data.time_range[1],
                                            '%Y-%m-%d').timestamp()
            center_time = start_time + (end_time - start_time) / 2
            delta = dt.timedelta(days=30)
            range_start = (dt.datetime.fromtimestamp(center_time) - delta).\
                strftime('%Y-%m-%d')
            range_end = (dt.datetime.fromtimestamp(center_time) + delta).\
                strftime('%Y-%m-%d')
            dataset = store.open_data(data_id=data.data_id,
                                      variable_names=variable_names,
                                      time_range=[range_start, range_end]
                                      )
            print(f'Attempting to normalize {data.data_id} ...')
            cube = normalize_dataset(dataset)
            try:
                print(f'Asserting {data.data_id} ...')
                assert_cube(cube)
                good_datasets.append(data.data_id)
            except ValueError as e:
                print(e)
                datasets_that_could_not_be_opened[data.data_id] = e
                continue
        print(f'Datasets with unsupported frequencies '
              f'(#{len(datasets_with_unsupported_frequencies)}): '
              f'{datasets_with_unsupported_frequencies}')
        print(f'Datasets without variables '
              f'(#{len(datasets_without_variables)}): '
              f'{datasets_without_variables}')
        print(f'Datasets that could not be opened '
              f'(#{len(datasets_that_could_not_be_opened)}): '
              f'{datasets_that_could_not_be_opened}')
        print(f'Datasets that were verified '
              f'(#{len(good_datasets)}): {good_datasets}')



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any
from typing import Mapping
from typing import Tuple
import geopandas as gpd
import pandas as pd
import shapely

from xcube_cci.cciodp import CciOdp
from xcube_cci.timerangegetter import TimeRangeGetter


class DataFrameAccessor:

    def __init__(self, cci_cdc: CciOdp, df_id: str, gdf_params: Mapping[str, Any]):
        self._cci_cdc = cci_cdc
        self._df_id = df_id
        self._metadata = cci_cdc.get_dataset_metadata(df_id)
        self._var_names = gdf_params.get("variable_names", {}).copy()
        if len(self._var_names) == 0:
            self._var_names = list(self._metadata.get("variable_infos").keys())
        self._spatial_subset_area = None
        if "bbox" in gdf_params:
            min_lon = gdf_params["bbox"][0]
            min_lat = gdf_params["bbox"][1]
            max_lon = gdf_params["bbox"][2]
            max_lat = gdf_params["bbox"][3]
            coords = ((min_lon, min_lat), (min_lon, max_lat), (max_lon, max_lat),
                      (max_lon, min_lat), (min_lon, min_lat))
            self._spatial_subset_area = shapely.Polygon(coords)
        # we use a time chunking of 1 as we do not know the actual content of a nc file
        self._time_chunking = 1
        tr = TimeRangeGetter(self._cci_cdc, self._metadata)
        self._time_ranges = tr.get_time_ranges(self._df_id, gdf_params)

    def request_time_range(self, time_index: int) -> Tuple:
        start_index = time_index * self._time_chunking
        end_index = ((time_index + 1) * self._time_chunking) - 1
        start_time = self._time_ranges[start_index][0]
        end_time = self._time_ranges[end_index][1]
        return start_time, end_time

    def get_geodataframe(self) -> gpd.GeoDataFrame:
        features = self._get_features()
        gdf = GeoDataFrame.from_features(features)
        return gdf

    def _get_features(self):
        for i in range(len(self._time_ranges)):
            return self._get_features_from_cci_cdc(i)

    def _get_features_from_cci_cdc(self, index: int = 0) -> gpd.GeoDataFrame:
        identifier = self._cci_cdc.get_dataset_id(self._df_id)
        start_time, end_time = self.request_time_range(time_index=index)
        try:
            start_time = start_time.tz_localize(None).isoformat()
            end_time = end_time.tz_localize(None).isoformat()
        except TypeError:
            # use unconverted time range values
            pass
        var_infos = self._metadata.get("variable_infos")
        all_var_names = list(var_infos.keys())
        var_names = self._var_names
        spatial_coord_names = ["lat", "lon", "latitude", "longitude", "geometry"]
        for spatial_coord_name in spatial_coord_names:
            if spatial_coord_name in all_var_names and \
                    spatial_coord_name not in var_names:
                var_names.append(spatial_coord_name)
        if "time" in all_var_names:
            var_names.append("time")
        if self._metadata.get("attributes", {}).get("shapefile", False):
            request = dict(parentIdentifier=identifier,
                           varNames=var_names,
                           startDate=start_time,
                           endDate=end_time,
                           drsId=self._df_id,
                           fileFormat='.shp')
            gdf = self._cci_cdc.get_geodataframe_from_shapefile(request)
        else:
            gdf_data = {}
            for var_name in var_names:
                if len(var_infos.get(var_name).get("dimensions", [])) > 1:
                    continue
                request = dict(parentIdentifier=identifier,
                               varNames=[var_name],
                               startDate=start_time,
                               endDate=end_time,
                               drsId=self._df_id,
                               fileFormat='.nc')
                gdf_data[var_name] = list(self._cci_cdc.get_data_chunk(
                    request, dim_indexes=(), to_bytes=False
                ))
            shift = False
            for lon_var in ["longitude", "lon"]:
                if lon_var in var_names:
                    if max(gdf_data[lon_var]) > 180.0:
                        shift = True
            if 'longitude' in var_names and 'latitude' in var_names:
                gdf = gpd.GeoDataFrame(
                    gdf_data, geometry=gpd.points_from_xy(
                        gdf_data['longitude'], gdf_data['latitude']
                    )
                )
                gdf = gdf.drop(['longitude', 'latitude'], axis=1)
            else:
                gdf = gpd.GeoDataFrame(
                    gdf_data,
                    geometry=gpd.points_from_xy(gdf_data['lon'], gdf_data['lat'])
                )
                gdf = gdf.drop(['lon', 'lat'], axis=1)
            if shift:
                gdf["geometry"] = gdf["geometry"].translate(xoff=-180.0)
        if self._spatial_subset_area:
            gdf["geometry"] = gdf.intersection(self._spatial_subset_area)
            gdf = gdf[~gdf["geometry"].is_empty]
        time_units = \
            self._metadata.get("variable_infos", {}).get("time", {}).get("units")
        if time_units and "since" in time_units:
            time_unit, t_offset = time_units.split("since")
            time_offset = pd.Timestamp(t_offset)
            time_unit = time_unit.strip().lower()

            def create_timestamp(df, tu, to):
                return to + pd.Timedelta(value=df['time'], unit=tu)

            gdf['time'] = gdf.apply(
                create_timestamp, args=(time_unit, time_offset), axis=1
            )
        if 'time' in gdf:
            gdf['time'] = gdf['time'].astype(str)
        else:
            gdf = gdf.assign(time=start_time)
        for feature in gdf.iterfeatures():
            yield feature


class GeoDataFrame:
    """
    Proxy for a ``geopandas.GeoDataFrame`` that holds an iterable of features
    or a feature collection for fastest possible streaming of GeoJSON features.

    Important note: although GeoDataFrame is not inherited from
    geopandas.GeoDataFrame it is an instance of that type meaning that::

        isinstance(GeoDataFrame(...), geopandas.GeoDataFrame) == True

    It also has all attributes of a geopandas.GeoDataFrame.
    This allows us using the proxy for operation parameters
    of type parameters pandas.DataFrame and geopandas.GeoDataFrame.
    """

    _OWN_PROPERTY_SET = {
        "_features",
        "features",
        "_lazy_data_frame",
        "lazy_data_frame",
        "close",
    }

    @classmethod
    def from_features(cls, features):
        """
        Create GeoDataFrame from an iterable of features
        or a feature collection.

        :param features: iterable of features or a feature collection,
            e.g., an open ``fiona.Collection`` instance.
        :return: An instance of a ``GeoDataFrame`` proxy.
        """
        return cls(features)

    def __init__(self, features):
        if features is None:
            raise ValueError('features must not be None')
        self._features = features
        self._lazy_data_frame = None

    @property
    def features(self):
        return self._features

    @property
    def lazy_data_frame(self):
        features = self._features
        if features is not None and self._lazy_data_frame is None:
            crs = features.crs if hasattr(features, 'crs') else None
            self._lazy_data_frame = gpd.GeoDataFrame.from_features(
                features, crs=crs
            )

        return self._lazy_data_frame

    def close(self):
        try:
            self._features.close()
        except AttributeError:
            pass
        self._features = None
        self._lazy_data_frame = None

    def __setattr__(self, key, value):
        if key in GeoDataFrame._OWN_PROPERTY_SET:
            object.__setattr__(self, key, value)
        else:
            self.lazy_data_frame.__setattr__(key, value)

    def __getattribute__(self, item):
        if item in GeoDataFrame._OWN_PROPERTY_SET:
            return object.__getattribute__(self, item)
        else:
            return getattr(self.lazy_data_frame, item)

    def __getattr__(self, item):
        if item in GeoDataFrame._OWN_PROPERTY_SET:
            return object.__getattribute__(self, item)
        else:
            return getattr(self.lazy_data_frame, item)

    def __getitem__(self, item):
        return self.lazy_data_frame.__getitem__(item)

    def __setitem__(self, key, value):
        return self.lazy_data_frame.__setitem__(key, value)

    def __str__(self):
        return str(self.lazy_data_frame)

    def __repr__(self):
        return repr(self.lazy_data_frame)

    def __len__(self):
        return len(self._features)
